"""
Tests for PKCS#11 functionality.

NOTE: these are not run in CI, due to lack of testing setup.
"""

import asyncio
import binascii
from io import BytesIO
from typing import Optional

import pytest
from asn1crypto import algos
from asn1crypto.algos import SignedDigestAlgorithm
from certomancer.registry import CertLabel
from freezegun import freeze_time
from pkcs11 import Mechanism, NoSuchKey, PKCS11Error
from pkcs11 import types as p11_types
from pyhanko.config.pkcs11 import PKCS11PinEntryMode, PKCS11SignatureConfig
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.sign import general, pkcs11, signers
from pyhanko.sign.general import SigningError
from pyhanko.sign.pkcs11 import (
    PKCS11Signer,
    PKCS11SigningContext,
    TokenCriteria,
    criteria_satisfied_by,
    find_token,
)

from pyhanko_certvalidator.registry import SimpleCertificateStore

from .samples import MINIMAL, TESTING_CA
from .signing_commons import (
    SIMPLE_DSA_V_CONTEXT,
    SIMPLE_ECC_V_CONTEXT,
    SIMPLE_ED448_V_CONTEXT,
    SIMPLE_ED25519_V_CONTEXT,
    SOFTHSM,
    async_val_trusted,
    pkcs11_only,
    pkcs11_test_module,
    val_trusted,
)

pytestmark = pkcs11_only


def _simple_sess(token='testrsa'):
    return pkcs11.open_pkcs11_session(
        pkcs11_test_module,
        user_pin='1234',
        token_criteria=TokenCriteria(label=token),
    )


default_other_certs = ('root', 'interm')
SIGNER_LABEL = 'signer1'


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@freeze_time('2020-11-01')
def test_simple_sign_with_rsassa_pss():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess(token='testrsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            prefer_pss=True,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    algo = emb.signer_info['signature_algorithm']['algorithm'].native
    assert algo == 'rsassa_pss'
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@freeze_time('2020-11-01')
def test_simple_sign_with_rsassa_pss_custom_parameters():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')

    pss_params = algos.RSASSAPSSParams(
        {
            'hash_algorithm': algos.DigestAlgorithm({'algorithm': 'sha256'}),
            'mask_gen_algorithm': algos.MaskGenAlgorithm(
                {
                    'algorithm': 'mgf1',
                    'parameters': algos.DigestAlgorithm(
                        {'algorithm': 'sha256'}
                    ),
                }
            ),
            'salt_length': 32,
        }
    )
    with _simple_sess(token='testrsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            prefer_pss=True,
            signature_mechanism=algos.SignedDigestAlgorithm(
                {'algorithm': 'rsassa_pss', 'parameters': pss_params},
            ),
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    algo = emb.signer_info['signature_algorithm']['algorithm'].native
    params = emb.signer_info['signature_algorithm']['parameters']
    assert algo == 'rsassa_pss'
    assert params.dump() == pss_params.dump()
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@freeze_time('2020-11-01')
def test_simple_sign_legacy_open_session_by_token_label():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with pytest.deprecated_call():
        with pkcs11.open_pkcs11_session(
            pkcs11_test_module, user_pin='1234', token_label='testrsa'
        ) as sess:
            signer = pkcs11.PKCS11Signer(
                sess,
                SIGNER_LABEL,
                other_certs_to_pull=default_other_certs,
            )
            out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_sign_external_certs(bulk_fetch):
    # Test to see if unnecessary fetches for intermediate certs are skipped

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            ca_chain=(TESTING_CA.get_cert(CertLabel('interm')),),
            bulk_fetch=bulk_fetch,
        )
        orig_fetcher = pkcs11._pull_cert
        try:

            def _trap_pull(session, *, label=None, cert_id=None):
                if label != SIGNER_LABEL:
                    raise RuntimeError
                return orig_fetcher(session, label=label, cert_id=cert_id)

            pkcs11._pull_cert = _trap_pull
            assert isinstance(signer.cert_registry, SimpleCertificateStore)
            assert len(list(signer.cert_registry)) == 1
            out = signers.sign_pdf(w, meta, signer=signer)
        finally:
            pkcs11._pull_cert = orig_fetcher

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_sign_multiple_cert_sources(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=('root',),
            ca_chain=(TESTING_CA.get_cert(CertLabel('interm')),),
            bulk_fetch=bulk_fetch,
        )
        assert isinstance(signer.cert_registry, SimpleCertificateStore)
        assert len(list(signer.cert_registry)) == 2
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_wrong_key_label(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
            key_label='NoSuchKeyExists',
        )
        with pytest.raises(NoSuchKey):
            signers.sign_pdf(w, meta, signer=signer)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_wrong_cert(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            key_label=SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
            cert_id=binascii.unhexlify(b'deadbeef'),
        )
        with pytest.raises(PKCS11Error, match='Could not find cert'):
            signers.sign_pdf(w, meta, signer=signer)


@freeze_time('2020-11-01')
def test_provided_certs():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            key_label=SIGNER_LABEL,
            signing_cert=signer_cert,
            ca_chain={
                TESTING_CA.get_cert(CertLabel('root')),
                TESTING_CA.get_cert(CertLabel('interm')),
            },
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    assert emb.signer_cert.dump() == signer_cert.dump()
    # this will fail if the intermediate cert is not present
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_signer_provided_others_pulled(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            ca_chain={
                TESTING_CA.get_cert(CertLabel('root')),
                TESTING_CA.get_cert(CertLabel('interm')),
            },
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_signer_pulled_others_provided(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            key_label=SIGNER_LABEL,
            signing_cert=signer_cert,
            bulk_fetch=bulk_fetch,
            other_certs_to_pull=default_other_certs,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    assert emb.signer_cert.dump() == signer_cert.dump()
    # this will fail if the intermediate cert is not present
    val_trusted(emb)


@freeze_time('2020-11-01')
def test_unclear_key_label_and_cert():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess() as sess:
        with pytest.raises(PKCS11Error, match='Found more than one'):
            signer = pkcs11.PKCS11Signer(sess)
            signers.sign_pdf(w, meta, signer=signer)


@freeze_time('2020-11-01')
def test_auto_use_only_key_if_cert_is_known():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    signer_cert = TESTING_CA.get_cert(CertLabel('signer1'))
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            signing_cert=signer_cert,
            other_certs_to_pull=default_other_certs,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign_dsa(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', md_algorithm='sha256'
    )
    with _simple_sess(token='testdsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb, vc=SIMPLE_DSA_V_CONTEXT())


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign_ecdsa(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', md_algorithm='sha256'
    )
    with _simple_sess(token='testecdsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
            use_raw_mechanism=True,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb, vc=SIMPLE_ECC_V_CONTEXT())


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign_ed25519(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess(token='tested25519') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb, vc=SIMPLE_ED25519_V_CONTEXT())


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign_ed448(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    with _simple_sess(token='tested448') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb, vc=SIMPLE_ED448_V_CONTEXT())


@freeze_time('2020-11-01')
def test_simple_sign_from_config():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria('testrsa'),
        cert_label=SIGNER_LABEL,
        user_pin='1234',
        other_certs_to_pull=None,
    )

    with PKCS11SigningContext(config) as signer:
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


def test_config_init_failure_signing_error():
    config = PKCS11SignatureConfig(
        module_path='.',
        token_criteria=TokenCriteria('testrsa'),
        cert_label=SIGNER_LABEL,
        user_pin='1234',
        other_certs_to_pull=None,
    )

    with pytest.raises(SigningError, match='error while opening session'):
        with PKCS11SigningContext(config):
            pass


@freeze_time('2020-11-01')
def test_sign_skip_login_fail():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria(label='testrsa'),
        cert_label=SIGNER_LABEL,
        prompt_pin=PKCS11PinEntryMode.SKIP,
    )

    # no key will be found, since we didn't bother logging in
    with pytest.raises(NoSuchKey):
        with PKCS11SigningContext(config) as signer:
            signers.sign_pdf(w, meta, signer=signer)


@pytest.mark.skipif(
    not SOFTHSM,
    reason=(
        "this test relies on SoftHSM not supporting the "
        "PROTECTED_AUTHENTICATION_PATH flag, and is disabled when running "
        "against other PKCS#11 implementations."
    ),
)
@freeze_time('2020-11-01')
def test_sign_deferred_auth():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria('testrsa'),
        cert_label=SIGNER_LABEL,
        prompt_pin=PKCS11PinEntryMode.DEFER,
    )

    # no key will be found, since we didn't bother logging in
    with pytest.raises(
        SigningError, match="Protected auth.*not supported by loaded module"
    ):
        with PKCS11SigningContext(config) as signer:
            signers.sign_pdf(w, meta, signer=signer)


@freeze_time('2020-11-01')
def test_simple_sign_with_raw_rsa():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria('testrsa'),
        cert_label=SIGNER_LABEL,
        user_pin='1234',
        other_certs_to_pull=None,
        raw_mechanism=True,
    )

    with PKCS11SigningContext(config) as signer:
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb)


@pytest.mark.parametrize('bulk_fetch', [True, False])
@freeze_time('2020-11-01')
def test_simple_sign_with_raw_dsa(bulk_fetch):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', md_algorithm='sha256'
    )
    with _simple_sess(token='testdsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
            use_raw_mechanism=True,
        )
        out = signers.sign_pdf(w, meta, signer=signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_trusted(emb, vc=SIMPLE_DSA_V_CONTEXT())


def test_no_raw_pss():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', md_algorithm='sha256'
    )
    with _simple_sess(token='testrsa') as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            use_raw_mechanism=True,
            prefer_pss=True,
        )
        with pytest.raises(NotImplementedError, match='PSS not available'):
            signers.sign_pdf(w, meta, signer=signer)


def test_unsupported_algo():
    with pytest.raises(NotImplementedError, match="2.999"):
        pkcs11.select_pkcs11_signing_params(
            algos.SignedDigestAlgorithm({'algorithm': '2.999'}),
            digest_algorithm='sha256',
            use_raw_mechanism=False,
        )


@pytest.mark.parametrize('md', ['sha256', 'sha384'])
def test_select_ecdsa_mech(md):
    # can't do a round-trip test since softhsm doesn't support these, but
    # we can at least verify that the selection works
    algo = f'{md}_ecdsa'
    result = pkcs11.select_pkcs11_signing_params(
        algos.SignedDigestAlgorithm({'algorithm': algo}),
        digest_algorithm=md,
        use_raw_mechanism=False,
    )
    assert result.sign_kwargs['mechanism'] == getattr(
        Mechanism, f"ECDSA_{md.upper()}"
    )


@pytest.mark.parametrize(
    'label,cert_id,no_results,exp_err',
    [
        (
            'foo',
            b'foo',
            True,
            "Could not find cert with label 'foo', ID '666f6f'.",
        ),
        ('foo', None, True, "Could not find cert with label 'foo'."),
        (None, b'foo', True, "Could not find cert with ID '666f6f'."),
        ('foo', None, False, "Found more than one cert with label 'foo'."),
    ],
)
def test_pull_err_fmt(label, cert_id, no_results, exp_err):
    err = pkcs11._format_pull_err_msg(
        no_results=no_results, label=label, cert_id=cert_id
    )
    assert err == exp_err


@pytest.mark.parametrize(
    'bulk_fetch,pss',
    [(True, True), (False, False), (True, False), (True, True)],
)
@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_simple_sign_from_config_async(bulk_fetch, pss):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria('testrsa'),
        other_certs_to_pull=default_other_certs,
        bulk_fetch=bulk_fetch,
        prefer_pss=pss,
        cert_label=SIGNER_LABEL,
        user_pin='1234',
    )
    async with PKCS11SigningContext(config=config) as signer:
        pdf_signer = signers.PdfSigner(meta, signer)
        out = await pdf_signer.async_sign_pdf(w)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    await async_val_trusted(emb)


@pytest.mark.skip  # FIXME flaky test, sometimes coredumps with SoftHSM
@pytest.mark.parametrize(
    'bulk_fetch,pss',
    [(True, True), (False, False), (True, False), (True, True)],
)
@pytest.mark.asyncio
async def test_async_sign_many_concurrent(bulk_fetch, pss):
    concurrent_count = 10
    config = PKCS11SignatureConfig(
        module_path=pkcs11_test_module,
        token_criteria=TokenCriteria(label='testrsa'),
        other_certs_to_pull=default_other_certs,
        bulk_fetch=bulk_fetch,
        prefer_pss=pss,
        cert_label=SIGNER_LABEL,
        user_pin='1234',
    )
    async with PKCS11SigningContext(config=config) as signer:

        async def _job(_i):
            w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
            meta = signers.PdfSignatureMetadata(
                field_name='Sig1', reason=f"PKCS#11 concurrency test #{_i}!"
            )
            pdf_signer = signers.PdfSigner(meta, signer)
            sig_result = await pdf_signer.async_sign_pdf(w, in_place=True)
            await asyncio.sleep(2)
            return _i, sig_result

        jobs = asyncio.as_completed(map(_job, range(1, concurrent_count + 1)))
        for finished_job in jobs:
            i, out = await finished_job
            r = PdfFileReader(out)
            emb = r.embedded_signatures[0]
            assert emb.field_name == 'Sig1'
            assert emb.sig_object['/Reason'].endswith(f"#{i}!")
            with freeze_time("2020-11-01"):
                await async_val_trusted(emb)


@pytest.mark.skip  # FIXME flaky test, sometimes coredumps with SoftHSM
@pytest.mark.parametrize(
    'bulk_fetch,pss',
    [(True, True), (False, False), (True, False), (True, True)],
)
@pytest.mark.asyncio
async def test_async_sign_raw_many_concurrent_no_preload_objs(bulk_fetch, pss):
    concurrent_count = 10

    # don't instantiate through PKCS11SigningContext
    # also, just sign raw strings, we want to exercise the correctness of
    # the awaiting logic in sign_raw for object loading
    with _simple_sess() as sess:
        signer = pkcs11.PKCS11Signer(
            sess,
            SIGNER_LABEL,
            other_certs_to_pull=default_other_certs,
            bulk_fetch=bulk_fetch,
        )

        async def _job(_i):
            payload = f"PKCS#11 concurrency test #{_i}!".encode('utf8')
            sig_result = await signer.async_sign_raw(payload, 'sha256')
            await asyncio.sleep(2)
            return _i, sig_result

        jobs = asyncio.as_completed(map(_job, range(1, concurrent_count + 1)))
        for finished_job in jobs:
            i, sig = await finished_job
            general.validate_raw(
                signature=sig,
                signed_data=f"PKCS#11 concurrency test #{i}!".encode('utf8'),
                cert=signer.signing_cert,
                md_algorithm='sha256',
                signature_algorithm=SignedDigestAlgorithm(
                    {'algorithm': 'sha256_rsa'}
                ),
            )


def test_token_does_not_exist():
    with pytest.raises(PKCS11Error, match='No token matching criteria'):
        _simple_sess(token='aintnosuchtoken')


def test_token_unclear():
    with pytest.raises(PKCS11Error, match='more than 1'):
        return pkcs11.open_pkcs11_session(
            pkcs11_test_module, user_pin='1234', token_label=None
        )


DUMMY_VER = {'major': 0, 'minor': 0}
DUMMY_ARGS = dict(
    slotDescription=b'',
    manufacturerID=b'',
    hardwareVersion=DUMMY_VER,
    firmwareVersion=DUMMY_VER,
)


class DummyToken(p11_types.Token):
    def open(self, rw=False, user_pin=None, so_pin=None):
        raise NotImplementedError


class DummySlot(p11_types.Slot):
    def __init__(self, lbl: Optional[str]):
        self.lbl = lbl

        super().__init__(
            "dummy.so.0",
            slot_id=0xDEADBEEF,
            flags=(
                p11_types.SlotFlag(0)
                if lbl is None
                else p11_types.SlotFlag.TOKEN_PRESENT
            ),
            **DUMMY_ARGS,
        )

    def get_token(self):
        if self.lbl is not None:
            lbl = self.lbl.encode('utf8')
            return DummyToken(
                self,
                label=lbl,
                model=b'DummyToken',
                flags=p11_types.TokenFlag(0),
                serialNumber=lbl + b'-\xde\xad\xbe\xef',
                **DUMMY_ARGS,
            )
        else:
            raise PKCS11Error("No token in slot")

    def get_mechanisms(self):
        return []

    def get_mechanism_info(self, mechanism):
        raise NotImplementedError


@pytest.mark.parametrize(
    'slot_list,slot_no_query,token_criteria',
    [
        (('foo',), None, None),
        (('foo', 'bar'), 0, TokenCriteria(label='foo')),
        (('foo', None, 'bar'), 0, TokenCriteria(label='foo')),
        (('foo', None, 'bar'), None, TokenCriteria(label='foo')),
        (
            ('foo', None, 'bar'),
            None,
            TokenCriteria(label='foo', serial=b'foo-\xde\xad\xbe\xef'),
        ),
        (
            ('foo', None, 'bar'),
            None,
            TokenCriteria(serial=b'foo-\xde\xad\xbe\xef'),
        ),
        (
            ('foo', None, 'bar'),
            None,
            TokenCriteria(serial=b'bar-\xde\xad\xbe\xef'),
        ),
        # skip over empty slots when doing this scan
        ((None, 'foo', None, 'bar'), None, TokenCriteria(label='foo')),
        ((None, 'foo', None), 1, None),
    ],
)
def test_find_token(slot_list, slot_no_query, token_criteria):
    tok = find_token(
        [DummySlot(lbl) for lbl in slot_list],
        slot_no=slot_no_query,
        token_criteria=token_criteria,
    )
    assert tok is not None
    criteria_satisfied_by(token_criteria, tok)


@pytest.mark.parametrize(
    'slot_list,slot_no_query,criteria,err',
    [
        (('foo', 'bar'), 2, TokenCriteria(label='foo'), 'too large'),
        (('foo', 'bar'), 1, TokenCriteria(label='foo'), 'label is not \'foo\''),
        (
            ('foo', 'bar'),
            1,
            TokenCriteria(serial=b'foo-\xde\xad\xbe\xef'),
            'serial is not \'666f6f2ddeadbeef\'',
        ),
        # when querying by slot, we want the error to be passed on
        ((None, 'bar'), 0, None, 'No token in'),
        (('foo', 'bar'), None, None, 'more than 1'),
        # right now, we don't care about the status of the slot in any way
        (('foo', None), None, None, 'more than 1'),
    ],
)
def test_find_token_error(slot_list, slot_no_query, criteria, err):
    with pytest.raises(PKCS11Error, match=err):
        find_token(
            [DummySlot(lbl) for lbl in slot_list],
            slot_no=slot_no_query,
            token_criteria=criteria,
        )


@pytest.mark.parametrize(
    'slot_list,token_lbl_query',
    [
        ((None, 'bar'), 'foo'),
        (('foo', 'bar'), 'baz'),
        ((None, None), 'foo'),
        ((), 'foo'),
    ],
)
def test_token_not_found(slot_list, token_lbl_query):
    tok = find_token(
        [DummySlot(lbl) for lbl in slot_list],
        slot_no=None,
        token_criteria=TokenCriteria(label=token_lbl_query),
    )
    assert tok is None


def _mirror(*cfgs):
    for cfg in cfgs:
        a = cfg
        b = cfg[1], cfg[0], cfg[3], cfg[2]
        yield a
        if a != b:
            yield b


@pytest.mark.parametrize(
    'key_in,cert_in,key_prop,cert_prop',
    [
        *_mirror(
            ((None, None), (None, None), (None, None), (None, None)),
            (('a', b"a"), (None, None), ('a', b"a"), ('a', b"a")),
            (('a', None), (None, None), ('a', None), ('a', None)),
            ((None, b"a"), (None, None), (None, b"a"), (None, b"a")),
            (('a', None), (None, b"a"), ('a', None), (None, b"a")),
        )
    ],
)
def test_config_fallbacks(key_in, cert_in, key_prop, cert_prop):
    # noinspection PyTypeChecker
    signer = PKCS11Signer(
        pkcs11_session=None,
        cert_label=cert_in[0],
        cert_id=cert_in[1],
        key_label=key_in[0],
        key_id=key_in[1],
    )
    actual_key_prop = signer.key_label, signer.key_id
    actual_cert_prop = signer.cert_label, signer.cert_id
    assert (actual_key_prop, actual_cert_prop) == (key_prop, cert_prop)



import asyncio
import hashlib
import os
from datetime import datetime
from io import BytesIO

import pyhanko.pdf_utils.content
import pyhanko.sign.fields
import pytest
from asn1crypto.algos import SignedDigestAlgorithm
from certomancer.integrations.illusionist import Illusionist
from certomancer.registry import CertLabel, KeyLabel
from freezegun import freeze_time
from pyhanko import stamp
from pyhanko.keys import load_cert_from_pemder, load_certs_from_pemder
from pyhanko.pdf_utils import embed, generic, layout
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
from pyhanko.pdf_utils.misc import PdfReadError
from pyhanko.pdf_utils.reader import PdfFileReader
from pyhanko.pdf_utils.writer import copy_into_new_writer
from pyhanko.sign import fields, signers, timestamps
from pyhanko.sign.diff_analysis import (
    NO_CHANGES_DIFF_POLICY,
    DiffResult,
    ModificationLevel,
)
from pyhanko.sign.general import SigningError, get_pyca_cryptography_hash
from pyhanko.sign.signers import cms_embedder
from pyhanko.sign.signers.pdf_byterange import BuildProps
from pyhanko.sign.signers.pdf_cms import (
    ExternalSigner,
    PdfCMSSignedAttributes,
    asyncify_signer,
)
from pyhanko.sign.signers.pdf_signer import PdfTBSDocument
from pyhanko.sign.validation import (
    DocumentSecurityStore,
    RevocationInfoValidationType,
    SignatureCoverageLevel,
    add_validation_info,
    apply_adobe_revocation_info,
    async_validate_detached_cms,
    read_certification_data,
    validate_pdf_ltv_signature,
    validate_pdf_signature,
    validate_pdf_timestamp,
)
from pyhanko.sign.validation.errors import SignatureValidationError
from pyhanko.stamp import QRStampStyle

from pyhanko_certvalidator import CertificateValidator, ValidationContext
from pyhanko_certvalidator.errors import PathValidationError
from pyhanko_certvalidator.registry import SimpleCertificateStore

from .samples import *
from .signing_commons import (
    DUMMY_HTTP_TS,
    DUMMY_TS,
    FIXED_OCSP,
    FROM_CA,
    FROM_CA_PKCS12,
    FROM_ECC_CA,
    FROM_ED448_CA,
    FROM_ED25519_CA,
    REVOKED_SIGNER,
    SELF_SIGN,
    SIMPLE_ED448_V_CONTEXT,
    SIMPLE_ED25519_V_CONTEXT,
    SIMPLE_V_CONTEXT,
    TRUST_ROOTS,
    TSA_CERT,
    async_val_trusted,
    dummy_ocsp_vc,
    live_testing_vc,
    val_trusted,
    val_untrusted,
)

DUMMY_TS_NO_NONCE = timestamps.DummyTimeStamper(
    tsa_cert=TSA_CERT,
    tsa_key=TESTING_CA.key_set.get_private_key('tsa'),
    certs_to_embed=FROM_CA.cert_registry,
    include_nonce=False,
)


def test_der_detect(tmp_path):
    from pathlib import Path

    tmp: Path = tmp_path / "test.der"
    orig_bytes = SELF_SIGN.signing_cert.dump()
    tmp.write_bytes(orig_bytes)
    result = load_cert_from_pemder(str(tmp))

    # make sure the resulting object gets parsed fully, for good measure
    # noinspection PyStatementEffect
    result.native
    assert result.dump() == orig_bytes


def test_enforce_one_cert():
    fname = CRYPTO_DATA_DIR + '/some-chain.cert.pem'

    assert len(list(load_certs_from_pemder([fname]))) == 2
    with pytest.raises(ValueError):
        load_cert_from_pemder(fname)


def test_simple_sign():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)


def test_simple_sign_tamper():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)

    # try tampering with the file
    out.seek(0x9D)
    # this just changes the size of the media box, so the file should remain
    # a valid PDF.
    out.write(b'4')
    out.seek(0)
    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    tampered = validate_pdf_signature(emb, SIMPLE_V_CONTEXT())
    assert not tampered.intact
    assert tampered.valid
    assert tampered.summary() == 'INVALID'


def test_simple_sign_fresh_doc():
    r = PdfFileReader(BytesIO(MINIMAL))
    w = copy_into_new_writer(r)
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)


def test_simple_sign_120mb_file():
    # put this in a function to avoid putting multiple copies
    #  of a huge buffer in locals
    def _gen_signed():
        w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
        w.root['/YugeObject'] = w.add_object(
            generic.StreamObject(stream_data=bytes(120 * 1024 * 1024))
        )
        w.update_root()

        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        return signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(_gen_signed())
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)


def test_append_sigfield_tu_on_signing():
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)
    spec = fields.SigFieldSpec(
        sig_field_name='Sig1',
        empty_field_appearance=True,
        readable_field_name="Test test",
    )
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.PdfSigner(
        meta, signer=SELF_SIGN, new_field_spec=spec
    ).sign_pdf(w)

    r = PdfFileReader(out)

    assert r.root['/AcroForm']['/Fields'][0]['/TU'] == "Test test"


@pytest.mark.parametrize(
    'policy, skip_diff',
    [(None, False), (NO_CHANGES_DIFF_POLICY, False), (None, True)],
)
def test_diff_fallback_ok(policy, skip_diff):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    status = validate_pdf_signature(
        emb, diff_policy=policy, skip_diff=skip_diff
    )
    if skip_diff:
        assert emb.diff_result is None
        # docmdp should still be OK without the diff check
        # because the signature covers the entire file
        assert status.docmdp_ok
        assert status.modification_level == ModificationLevel.NONE
    else:
        assert isinstance(emb.diff_result, DiffResult)
        assert status.modification_level == ModificationLevel.NONE
        assert status.docmdp_ok


def test_no_diff_summary():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    # just do an incremental DSS update
    DocumentSecurityStore.add_dss(
        out, sig_contents=None, certs=(SELF_SIGN.signing_cert,)
    )

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    status = validate_pdf_signature(emb, skip_diff=True)
    assert emb.diff_result is None
    assert status.modification_level is None
    assert not status.docmdp_ok
    assert status.coverage == SignatureCoverageLevel.ENTIRE_REVISION
    assert 'EXTENDED' in status.summary()


@freeze_time('2020-11-01')
def test_sign_with_trust():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    assert '/AP' not in s.sig_field
    status = val_untrusted(s)
    assert not status.trusted

    val_trusted(s)


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_with_trust_async():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = await signers.async_sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    assert '/AP' not in s.sig_field
    await async_val_trusted(s)


@freeze_time('2020-12-05')
def test_sign_with_revoked(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1'),
        signer=REVOKED_SIGNER,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]

    vc = live_testing_vc(requests_mock)
    val_status = validate_pdf_signature(s, vc)
    assert val_status.intact
    assert val_status.valid
    assert val_status.revoked
    assert not val_status.trusted
    assert 'revoked' in val_status.pretty_print_details()
    summ = val_status.summary()
    assert 'INTACT' in summ
    assert 'REVOKED' in summ
    assert val_status.coverage == SignatureCoverageLevel.ENTIRE_FILE
    assert val_status.modification_level == ModificationLevel.NONE
    assert not val_status.bottom_line

    # should refuse to sign with a known revoked cert
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with pytest.raises(SigningError):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1', validation_context=vc
            ),
            signer=REVOKED_SIGNER,
        )


def test_sign_with_later_revoked_nots(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with freeze_time('2020-01-20'):
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=REVOKED_SIGNER,
        )
        r = PdfFileReader(out)
        s = r.embedded_signatures[0]

    # there's no way to do a timestamp validation check here, so the checker
    # should assume the timestamp to be invalid
    with freeze_time('2020-12-05'):
        r = PdfFileReader(out)
        s = r.embedded_signatures[0]
        vc = live_testing_vc(requests_mock)
        val_status = validate_pdf_signature(s, vc)
        assert val_status.intact
        assert val_status.valid
        assert val_status.revoked
        assert not val_status.trusted

        summ = val_status.summary()
        assert 'INTACT' in summ
        assert 'REVOKED' in summ
        assert val_status.coverage == SignatureCoverageLevel.ENTIRE_FILE
        assert val_status.modification_level == ModificationLevel.NONE
        assert not val_status.bottom_line


@freeze_time('2020-11-01')
def test_sign_with_trust_pkcs12():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1'),
        signer=FROM_CA_PKCS12,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    status = val_untrusted(s)
    assert not status.trusted

    val_trusted(s)


sign_test_files = (MINIMAL, MINIMAL_ONE_FIELD)


def test_enumerate_empty():
    with pytest.raises(StopIteration):
        next(fields.enumerate_sig_fields(PdfFileReader(BytesIO(MINIMAL))))


def test_certify_blank():
    r = PdfFileReader(BytesIO(MINIMAL))
    assert read_certification_data(r) is None


@freeze_time('2020-11-01')
def test_certify():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            certify=True,
            docmdp_permissions=pyhanko.sign.fields.MDPPerm.NO_CHANGES,
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    refs = s.sig_object.get_object()['/Reference']
    assert len(refs) == 1
    assert s.field_name == 'Sig1'
    val_trusted(s)

    info = read_certification_data(r)
    assert info.author_sig == s.sig_object.get_object()
    assert info.permission == pyhanko.sign.fields.MDPPerm.NO_CHANGES

    # with NO_CHANGES, we shouldn't be able to append an approval signature
    out.seek(0)
    w = IncrementalPdfFileWriter(out)
    with pytest.raises(SigningError):
        signers.sign_pdf(
            w, signers.PdfSignatureMetadata(field_name='Sig2'), signer=FROM_CA
        )


@freeze_time('2020-11-01')
def test_no_certify_after_sign():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    val_trusted(s)

    out.seek(0)
    w = IncrementalPdfFileWriter(out)
    with pytest.raises(SigningError, match='must be the first'):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig2',
                certify=True,
                docmdp_permissions=pyhanko.sign.fields.MDPPerm.FILL_FORMS,
            ),
            signer=FROM_CA,
        )


@freeze_time('2020-11-01')
def test_approval_sig():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            certify=True,
        ),
        signer=FROM_CA,
    )
    out.seek(0)
    w = IncrementalPdfFileWriter(out)
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig2'), signer=FROM_CA
    )

    out.seek(0)

    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    val_trusted(s, extd=True)

    info = read_certification_data(r)
    assert info.author_sig == s.sig_object.get_object()
    assert info.permission == pyhanko.sign.fields.MDPPerm.FILL_FORMS

    s = r.embedded_signatures[1]
    assert s.field_name == 'Sig2'
    val_trusted(s)


@freeze_time('2020-11-01')
def test_ocsp_embed():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            embed_validation_info=True,
        ),
        signer=FROM_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    status = val_untrusted(s)
    assert not status.trusted

    val_trusted(s)

    vc = apply_adobe_revocation_info(s.signer_info)
    assert len(vc.ocsps) == 1


@freeze_time('2020-11-01')
def test_ocsp_without_nextupdate_embed(requests_mock):
    ca = CERTOMANCER.get_pki_arch(ArchLabel('testing-ca-ocsp-no-nextupdate'))
    vc = ValidationContext(
        trust_roots=[ca.get_cert(CertLabel('root'))],
        allow_fetching=True,
        other_certs=[],
    )

    signer = signers.SimpleSigner(
        signing_cert=ca.get_cert(CertLabel('signer-special')),
        signing_key=ca.key_set.get_private_key(KeyLabel('signer1')),
        cert_registry=SimpleCertificateStore.from_certs(
            [ca.get_cert(CertLabel('root')), ca.get_cert(CertLabel('interm'))]
        ),
    )
    Illusionist(ca).register(requests_mock)
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            embed_validation_info=True,
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    vc = apply_adobe_revocation_info(s.signer_info)
    assert len(vc.ocsps) == 1
    assert len(vc.crls) == 1

    simple_response = vc.ocsps[0]['response_bytes']['response']
    rdata = simple_response.parsed['tbs_response_data']
    assert rdata['responses'][0]['next_update'].native is None


@freeze_time('2020-11-01')
def test_adobe_revinfo_live(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    vc = live_testing_vc(requests_mock)
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=vc,
            subfilter=fields.SigSeedSubFilter.ADOBE_PKCS7_DETACHED,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    rivt_adobe = RevocationInfoValidationType.ADOBE_STYLE
    status = validate_pdf_ltv_signature(
        r.embedded_signatures[0], rivt_adobe, {'trust_roots': TRUST_ROOTS}
    )
    assert status.valid and status.trusted


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_meta_tsa_verify():
    # check if my testing setup works
    vc = ValidationContext(
        trust_roots=TRUST_ROOTS,
        allow_fetching=False,
        crls=[],
        ocsps=[FIXED_OCSP],
        revocation_mode='hard-fail',
    )
    with pytest.raises(PathValidationError):
        cv = CertificateValidator(TSA_CERT, validation_context=vc)
        await cv.async_validate_usage({'time_stamping'})


@freeze_time('2020-11-01')
def test_adobe_revinfo_live_nofullchain():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            validation_context=dummy_ocsp_vc(),
            subfilter=fields.SigSeedSubFilter.ADOBE_PKCS7_DETACHED,
            embed_validation_info=True,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
    )
    r = PdfFileReader(out)
    rivt_adobe = RevocationInfoValidationType.ADOBE_STYLE
    # same as for the pades test above
    with pytest.raises(SignatureValidationError):
        validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_adobe,
            {
                'trust_roots': TRUST_ROOTS,
                'allow_fetching': False,
                'ocsps': [FIXED_OCSP],
            },
        )
    from requests_mock import Mocker

    with Mocker() as m:
        live_testing_vc(m)
        status = validate_pdf_ltv_signature(
            r.embedded_signatures[0],
            rivt_adobe,
            {'trust_roots': TRUST_ROOTS, 'allow_fetching': True},
        )
        assert status.valid and not status.trusted, status.summary()


@freeze_time('2020-11-01')
def test_simple_qr_sign():
    style = QRStampStyle(stamp_text="Hi, it's\n%(ts)s")
    signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='Sig1'),
        FROM_CA,
        stamp_style=style,
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = signer.sign_pdf(
        w,
        existing_fields_only=True,
        appearance_text_params={'url': 'https://example.com'},
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    assert '/QR' in s.sig_field['/AP']['/N']['/Resources']['/XObject']

    val_trusted(s)


@pytest.mark.parametrize('params_value', [None, {}, {'some': 'value'}])
def test_qr_sign_enforce_url_param(params_value):
    style = QRStampStyle(stamp_text="Hi, it's\n%(ts)s")
    signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='Sig1'),
        FROM_CA,
        stamp_style=style,
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    with pytest.raises(layout.LayoutError):
        signer.sign_pdf(
            w, existing_fields_only=True, appearance_text_params=params_value
        )


def test_bytes_reserved_even():
    with pytest.raises(ValueError):
        signers.PdfByteRangeDigest(bytes_reserved=1)


def test_name_location():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))

    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', name='Bleh', location='Bluh'
    )
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)

    assert emb.sig_object['/Name'] == 'Bleh'
    assert emb.sig_object['/Location'] == 'Bluh'


def test_no_email():
    # just sign with any cert, don't care about validation etc.
    # This is simply to test the name generation logic if no email address
    # is available
    signer = signers.SimpleSigner.load(
        CRYPTO_DATA_DIR + '/keys-rsa/tsa.key.pem',
        CRYPTO_DATA_DIR + '/tsa.cert.pem',
        ca_chain_files=(),
        key_passphrase=b'secret',
    )

    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
    )
    pdf_signer = signers.PdfSigner(
        meta,
        signer=signer,
        stamp_style=stamp.TextStampStyle(
            stamp_text='%(signer)s\n%(ts)s',
        ),
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))
    out = pdf_signer.sign_pdf(
        w,
    )

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)

    ap_data = emb.sig_field['/AP']['/N'].data
    cn = signer.signing_cert.subject.native['common_name'].encode('ascii')
    assert cn in ap_data


def _tamper_with_sig_obj(tamper_fun):
    input_buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(input_buf)
    md_algorithm = 'sha256'

    cms_writer = cms_embedder.PdfCMSEmbedder().write_cms(
        field_name='Signature', writer=w
    )
    next(cms_writer)
    sig_obj = signers.SignatureObject(bytes_reserved=8192)

    cms_writer.send(cms_embedder.SigObjSetup(sig_placeholder=sig_obj))

    tamper_fun(w, sig_obj)

    prep_document_hash, output = cms_writer.send(
        cms_embedder.SigIOSetup(md_algorithm=md_algorithm, in_place=True)
    )

    signer: signers.SimpleSigner = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        cms_obj = signer.sign(
            data_digest=prep_document_hash.document_digest,
            digest_algorithm=md_algorithm,
        )
    cms_writer.send(cms_obj)
    return output


@freeze_time('2020-11-01')
def test_sig_delete_type():
    # test whether deleting /Type defaults to /Sig
    def tamper(writer, sig_obj):
        del sig_obj['/Type']

    out = _tamper_with_sig_obj(tamper)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    val_trusted(emb)

    # ... but the doctimestamp validator shouldn't let that slide
    # (yes, obviously this also isn't a valid timestamp token, hence the
    #  match=... rule here)
    with pytest.raises(
        SignatureValidationError, match='.*must be /DocTimeStamp.*'
    ):
        validate_pdf_timestamp(emb, validation_context=SIMPLE_V_CONTEXT())


def test_timestamp_wrong_type():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    # Again:
    # (yes, obviously this also isn't a valid timestamp token, hence the
    #  match=... rule here)
    with pytest.raises(
        SignatureValidationError, match='.*must be /DocTimeStamp.*'
    ):
        validate_pdf_timestamp(emb, validation_context=SIMPLE_V_CONTEXT())


@pytest.mark.parametrize(
    'wrong_subfilter',
    [pdf_name('/abcde'), pdf_name("/ETSI.RFC3161"), None, generic.NullObject()],
)
@freeze_time('2020-11-01')
def test_sig_wrong_subfilter(wrong_subfilter):
    def tamper(writer, sig_obj):
        if wrong_subfilter:
            sig_obj['/SubFilter'] = wrong_subfilter
        else:
            del sig_obj['/SubFilter']

    out = _tamper_with_sig_obj(tamper)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    with pytest.raises(SignatureValidationError):
        val_trusted(emb)


@freeze_time('2020-11-01')
def test_sig_no_contents():
    def tamper(writer, sig_obj):
        # the placeholder object needs to be written to, to make the
        # run its course
        sig_obj['/FakeContents'] = sig_obj['/Contents']
        del sig_obj['/Contents']

    out = _tamper_with_sig_obj(tamper)

    r = PdfFileReader(out)
    with pytest.raises(PdfReadError, match='.*Could not read /Contents.*'):
        # noinspection PyStatementEffect
        r.embedded_signatures[0]


@freeze_time('2020-11-01')
def test_sig_null_contents():
    def tamper(writer, sig_obj):
        sig_obj['/FakeContents'] = sig_obj['/Contents']
        sig_obj['/Contents'] = generic.NullObject()

    out = _tamper_with_sig_obj(tamper)

    r = PdfFileReader(out)
    with pytest.raises(PdfReadError, match='.*string-like.*'):
        # noinspection PyStatementEffect
        r.embedded_signatures[0]


@freeze_time('2020-11-01')
def test_sig_indirect_contents():
    def tamper(writer, sig_obj):
        sig_obj['/Contents'] = writer.add_object(sig_obj['/Contents'])

    out = _tamper_with_sig_obj(tamper)

    r = PdfFileReader(out)
    with pytest.raises(PdfReadError, match='.*be string-like.*'):
        # noinspection PyStatementEffect
        r.embedded_signatures[0]


@freeze_time('2020-11-01')
def test_timestamp_with_different_digest():
    ts = timestamps.DummyTimeStamper(
        tsa_cert=TSA_CERT,
        tsa_key=TESTING_CA.key_set.get_private_key('tsa'),
        certs_to_embed=FROM_CA.cert_registry,
        override_md='sha512',
    )
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL_ONE_FIELD))

    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(md_algorithm='sha256'),
        signer=FROM_CA,
        timestamper=ts,
        existing_fields_only=True,
    )

    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    assert s.field_name == 'Sig1'
    validity = val_trusted(s)
    assert validity.timestamp_validity is not None
    assert validity.timestamp_validity.trusted


def test_sign_with_empty_kids():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    fields.append_signature_field(
        w,
        fields.SigFieldSpec(
            sig_field_name='Sig1',
            combine_annotation=False,
            box=(20, 20, 80, 40),
        ),
    )
    w.root['/AcroForm']['/Fields'][0]['/Kids'] = generic.ArrayObject()
    meta = signers.PdfSignatureMetadata(field_name='Sig1')

    with pytest.raises(SigningError, match="Failed to access.*annot.*"):
        signers.sign_pdf(w, meta, signer=FROM_CA)


@freeze_time('2020-11-01')
def test_sign_without_annot():
    with open(PDF_DATA_DIR + '/minimal-annotless.pdf', 'rb') as f:
        w = IncrementalPdfFileWriter(f)
        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        out = signers.sign_pdf(w, meta, signer=FROM_CA)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    assert '/AP' not in emb.sig_field
    assert '/Rect' not in emb.sig_field
    assert '/Kids' not in emb.sig_field
    assert '/Type' not in emb.sig_field
    val_trusted(emb)


@pytest.mark.parametrize('in_place', [True, False])
@freeze_time('2020-11-01')
def test_no_revinfo_to_be_added(requests_mock, in_place):
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)

    vc = live_testing_vc(requests_mock)
    signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            embed_validation_info=True,
            validation_context=vc,
            subfilter=fields.SigSeedSubFilter.PADES,
        ),
        signer=FROM_CA,
        timestamper=DUMMY_TS,
        in_place=True,
    )

    orig_file_length = buf.seek(0, os.SEEK_END)
    r = PdfFileReader(buf)
    emb_sig = r.embedded_signatures[0]
    orig_dss = DocumentSecurityStore.read_dss(r)
    assert len(orig_dss.ocsps) == 1
    assert len(orig_dss.crls) == 1
    # test with same vc, this shouldn't change anything
    # Turn off VRI updates, since those always trigger a write.
    output = add_validation_info(
        emb_sig, vc, in_place=in_place, add_vri_entry=False
    )
    if in_place:
        assert output is r.stream

    new_file_length = output.seek(0, os.SEEK_END)
    assert orig_file_length == new_file_length
    new_dss = DocumentSecurityStore.read_dss(PdfFileReader(output))
    assert len(new_dss.ocsps) == 1
    assert len(new_dss.crls) == 1


@pytest.mark.parametrize('with_vri', [True, False])
def test_add_revinfo_timestamp_separate_no_dss(requests_mock, with_vri):
    buf = BytesIO(MINIMAL)
    w = IncrementalPdfFileWriter(buf)

    # create signature & timestamp without revocation info
    with freeze_time('2020-11-01'):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=FROM_CA,
            in_place=True,
        )
        signers.PdfTimeStamper(timestamper=DUMMY_TS).timestamp_pdf(
            IncrementalPdfFileWriter(buf), 'sha256', in_place=True
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]
        add_validation_info(emb_sig, vc, in_place=True, add_vri_entry=with_vri)

        r = PdfFileReader(buf)
        emb_sig = r.embedded_signatures[0]

        # without retroactive revinfo, the validation should fail
        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS},
        )
        assert status.valid and not status.trusted

        # with retroactive revinfo, it should be OK
        status = validate_pdf_ltv_signature(
            emb_sig,
            RevocationInfoValidationType.PADES_LT,
            {'trust_roots': TRUST_ROOTS, 'retroactive_revinfo': True},
        )
        assert status.valid and status.trusted
        assert status.modification_level == ModificationLevel.LTA_UPDATES


def test_add_revinfo_without_timestamp(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))

    # create signature without revocation info
    with freeze_time('2020-11-01'):
        out = signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(field_name='Sig1'),
            signer=FROM_CA,
            in_place=True,
        )

    # fast forward 1 month
    with freeze_time('2020-12-01'):
        vc = live_testing_vc(requests_mock)
        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]
        out = add_validation_info(emb_sig, vc)

        r = PdfFileReader(out)
        emb_sig = r.embedded_signatures[0]

        # even with revinfo, this should fail for lack of a timestamp
        with pytest.raises(
            SignatureValidationError, match='.*trusted timestamp.*'
        ):
            validate_pdf_ltv_signature(
                emb_sig,
                RevocationInfoValidationType.PADES_LT,
                {'trust_roots': TRUST_ROOTS, 'retroactive_revinfo': True},
            )

        # ... and certainly for LTA
        with pytest.raises(SignatureValidationError, match='Purported.*LTA.*'):
            validate_pdf_ltv_signature(
                emb_sig,
                RevocationInfoValidationType.PADES_LTA,
                {'trust_roots': TRUST_ROOTS, 'retroactive_revinfo': True},
            )


@pytest.mark.asyncio
async def test_embed_signed_attachment():
    dt = datetime.fromisoformat('2020-11-01T05:00:00+00:00')
    signature = await FROM_CA.async_sign_general_data(
        VECTOR_IMAGE_PDF, 'sha256', PdfCMSSignedAttributes(signing_time=dt)
    )

    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    signers.embed_payload_with_cms(
        w,
        file_spec_string='attachment.pdf',
        payload=embed.EmbeddedFileObject.from_file_data(
            w,
            data=VECTOR_IMAGE_PDF,
            mime_type='application/pdf',
            params=embed.EmbeddedFileParams(
                creation_date=dt, modification_date=dt
            ),
        ),
        cms_obj=signature,
        file_name='添付ファイル.pdf',
        file_spec_kwargs={'description': "Signed attachment test"},
    )
    out = BytesIO()
    w.write(out)

    r = PdfFileReader(out)
    emb_lst = r.root['/Names']['/EmbeddedFiles']['/Names']
    assert len(emb_lst) == 4
    assert emb_lst[0] == 'attachment.pdf'
    spec_obj = emb_lst[1]
    assert spec_obj['/UF'] == '添付ファイル.pdf'
    stream = spec_obj['/EF']['/F']
    assert stream.data == VECTOR_IMAGE_PDF
    assert spec_obj['/RF']['/F'][0] == 'attachment.sig'
    assert spec_obj['/RF']['/UF'][0] == '添付ファイル.sig'
    rel_file_ref = spec_obj['/RF']['/F'].raw_get(1).reference

    assert emb_lst[2] == 'attachment.sig'
    spec_obj = emb_lst[3]
    assert spec_obj['/UF'] == '添付ファイル.sig'
    stream = spec_obj['/EF']['/F']
    assert stream.data == signature.dump()
    assert stream.container_ref == rel_file_ref


@freeze_time('2020-11-01')
def test_simple_interrupted_signature():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='SigNew'), signer=FROM_CA
    )
    with pytest.deprecated_call():
        # noinspection PyDeprecation
        prep_digest, tbs_document, output = pdf_signer.digest_doc_for_signing(w)
    md_algorithm = tbs_document.md_algorithm
    assert tbs_document.post_sign_instructions is None

    # copy the output to a new buffer, just to make a point
    new_output = BytesIO()
    assert isinstance(output, BytesIO)
    buf = output.getbuffer()
    new_output.write(buf)
    buf.release()

    with pytest.deprecated_call():
        # noinspection PyDeprecation
        PdfTBSDocument.finish_signing(
            new_output,
            prep_digest,
            FROM_CA.sign(
                prep_digest.document_digest,
                digest_algorithm=md_algorithm,
            ),
        )

    r = PdfFileReader(new_output)
    val_trusted(r.embedded_signatures[0])


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_interrupted_with_delayed_signing_cert_mech():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='SigNew'),
        signer=ExternalSigner(
            signing_cert=None,
            cert_registry=None,
            signature_value=256,
            signature_mechanism=SignedDigestAlgorithm(
                {'algorithm': 'sha256_rsa'}
            ),
        ),
    )
    (
        prep_digest,
        tbs_document,
        output,
    ) = await pdf_signer.async_digest_doc_for_signing(w, bytes_reserved=8192)
    md_algorithm = tbs_document.md_algorithm
    assert tbs_document.post_sign_instructions is None

    # copy the output to a new buffer, just to make a point
    new_output = BytesIO()
    assert isinstance(output, BytesIO)
    buf = output.getbuffer()
    new_output.write(buf)
    buf.release()

    await PdfTBSDocument.async_finish_signing(
        new_output,
        prep_digest,
        await FROM_CA.async_sign(
            prep_digest.document_digest,
            digest_algorithm=md_algorithm,
        ),
    )

    r = PdfFileReader(new_output)
    await async_val_trusted(r.embedded_signatures[0])


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_interrupted_with_delayed_signing_cert_directly_specify_md():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(
            field_name='SigNew', md_algorithm='sha256'
        ),
        signer=ExternalSigner(
            signing_cert=None,
            cert_registry=None,
            signature_value=256,
        ),
    )
    (
        prep_digest,
        tbs_document,
        output,
    ) = await pdf_signer.async_digest_doc_for_signing(w, bytes_reserved=8192)
    md_algorithm = tbs_document.md_algorithm
    assert tbs_document.post_sign_instructions is None

    await PdfTBSDocument.async_finish_signing(
        output,
        prep_digest,
        await FROM_CA.async_sign(
            prep_digest.document_digest,
            digest_algorithm=md_algorithm,
        ),
    )

    r = PdfFileReader(output)
    await async_val_trusted(r.embedded_signatures[0])


@pytest.mark.asyncio
async def test_interrupted_could_not_determine_digest_algo():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='SigNew'),
        signer=ExternalSigner(
            signing_cert=None, cert_registry=None, signature_value=256
        ),
    )

    with pytest.raises(
        SigningError, match="Could not select.*digest algorithm"
    ):
        await pdf_signer.async_digest_doc_for_signing(w)


@pytest.mark.asyncio
async def test_interrupted_no_estimation():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(field_name='SigNew'),
        signer=ExternalSigner(
            signing_cert=None,
            cert_registry=None,
            signature_value=bytes(256),
            signature_mechanism=SignedDigestAlgorithm(
                {'algorithm': 'sha256_rsa'}
            ),
        ),
    )

    with pytest.raises(SigningError, match="estimation.*bytes_reserved"):
        await pdf_signer.async_digest_doc_for_signing(w)


@pytest.mark.asyncio
async def test_interrupted_with_delayed_signing_no_prevalidation():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    pdf_signer = signers.PdfSigner(
        signers.PdfSignatureMetadata(
            field_name='SigNew',
            embed_validation_info=True,
            validation_context=SIMPLE_V_CONTEXT(),
        ),
        signer=ExternalSigner(
            signing_cert=None,
            cert_registry=None,
            signature_value=256,
            signature_mechanism=SignedDigestAlgorithm(
                {'algorithm': 'sha256_rsa'}
            ),
        ),
    )
    with pytest.raises(SigningError, match='certificate must be provided'):
        await pdf_signer.async_digest_doc_for_signing(w, bytes_reserved=8192)


def test_determine_mechanism_no_signing_cert():
    signer = ExternalSigner(
        signing_cert=None, cert_registry=None, signature_value=bytes(256)
    )

    with pytest.raises(SigningError, match="Could not set up.*mechanism"):
        signer.get_signature_mechanism_for_digest('sha256')


@pytest.mark.asyncio
async def test_signer_info_no_signing_cert():
    signer = ExternalSigner(
        signing_cert=None,
        cert_registry=None,
        signature_value=256,
        signature_mechanism=SignedDigestAlgorithm({'algorithm': 'sha256_rsa'}),
    )

    with pytest.raises(
        SigningError, match="certificate must be available.*SignerInfo"
    ):
        await signer.async_sign_general_data(
            b"Hello world",
            digest_algorithm="sha256",
        )


@freeze_time('2020-11-01')
@pytest.mark.asyncio
async def test_sign_prescribed_attrs(requests_mock):
    vc = live_testing_vc(requests_mock)
    message = b'Hello world!'
    digest = hashlib.sha256(message).digest()
    signed_attrs = await FROM_CA.signed_attrs(digest, 'sha256')
    sig_cms = await FROM_CA.async_sign_prescribed_attributes(
        'sha256', signed_attrs=signed_attrs, timestamper=DUMMY_HTTP_TS
    )
    status = await async_validate_detached_cms(
        b'Hello world!', sig_cms['content'], signer_validation_context=vc
    )
    assert status.valid and status.intact and status.trusted
    ts_status = status.timestamp_validity
    assert ts_status.valid and ts_status.intact and ts_status.trusted


# noinspection PyDeprecation
@freeze_time('2020-11-01')
def test_sign_prescribed_attrs_legacy(requests_mock):
    vc = live_testing_vc(requests_mock)
    message = b'Hello world!'
    digest = hashlib.sha256(message).digest()
    signed_attrs = asyncio.run(FROM_CA.signed_attrs(digest, 'sha256'))
    with pytest.deprecated_call():
        sig_cms = FROM_CA.sign_prescribed_attributes(
            'sha256', signed_attrs=signed_attrs, timestamper=DUMMY_HTTP_TS
        )

    from pyhanko.sign.validation import validate_detached_cms

    with pytest.deprecated_call():
        status = validate_detached_cms(
            b'Hello world!', sig_cms['content'], signer_validation_context=vc
        )
    assert status.valid and status.intact and status.trusted
    ts_status = status.timestamp_validity
    assert ts_status.valid and ts_status.intact and ts_status.trusted


@freeze_time('2020-11-01')
def test_sign_tight_container():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', tight_size_estimates=True
    )
    out = signers.sign_pdf(w, meta, signer=SELF_SIGN)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)

    contents_str = emb.pkcs7_content
    ci = cms.ContentInfo(
        {'content_type': 'signed_data', 'content': emb.signed_data}
    )
    assert ci.dump() == contents_str


@freeze_time('2020-11-01')
def test_sign_tight_container_with_ts():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', tight_size_estimates=True, md_algorithm='sha256'
    )
    out = signers.sign_pdf(
        w, meta, signer=SELF_SIGN, timestamper=DUMMY_TS_NO_NONCE
    )

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)

    contents_str = emb.pkcs7_content
    ci = cms.ContentInfo(
        {'content_type': 'signed_data', 'content': emb.signed_data}
    )
    assert ci.dump() == contents_str


@freeze_time('2020-11-01')
def test_sign_tight_container_with_lta(requests_mock):
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        tight_size_estimates=True,
        subfilter=fields.SigSeedSubFilter.PADES,
        use_pades_lta=True,
        embed_validation_info=True,
        validation_context=live_testing_vc(requests_mock),
    )
    out = signers.sign_pdf(
        w,
        meta,
        signer=FROM_CA,
        timestamper=DUMMY_TS_NO_NONCE,
    )

    r = PdfFileReader(out)

    def _check(emb):
        contents_str = emb.pkcs7_content
        ci = cms.ContentInfo(
            {'content_type': 'signed_data', 'content': emb.signed_data}
        )
        assert ci.dump() == contents_str

    _check(r.embedded_regular_signatures[0])
    _check(r.embedded_timestamp_signatures[0])


# noinspection PyAbstractClass
@asyncify_signer
class LegacyRSASigner(signers.Signer):
    def __init__(
        self,
        signing_cert,
        signing_key,
        cert_registry,
        signature_mechanism: SignedDigestAlgorithm = None,
        prefer_pss=False,
    ):
        self.signing_key = signing_key
        super().__init__(
            prefer_pss=prefer_pss,
            signing_cert=signing_cert,
            cert_registry=cert_registry,
            signature_mechanism=signature_mechanism,
        )

    # noinspection PyUnusedLocal
    def sign_raw(
        self, data: bytes, digest_algorithm: str, dry_run=False
    ) -> bytes:
        from cryptography.hazmat.primitives import serialization

        priv_key = serialization.load_der_private_key(
            self.signing_key.dump(), password=None
        )

        from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15

        padding = PKCS1v15()
        hash_algo = get_pyca_cryptography_hash(digest_algorithm)
        return priv_key.sign(data, padding, hash_algo)


def test_simple_sign_legacy_signer_upgrade():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1')
    legacy_signer = LegacyRSASigner(
        signing_cert=SELF_SIGN.signing_cert,
        signing_key=SELF_SIGN.signing_key,
        cert_registry=SELF_SIGN.cert_registry,
    )
    out = signers.sign_pdf(w, meta, signer=legacy_signer)

    r = PdfFileReader(out)
    emb = r.embedded_signatures[0]
    assert emb.field_name == 'Sig1'
    val_untrusted(emb)


def test_disallow_hybrid_sign():
    fname = 'minimal-hybrid-xref.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        w = IncrementalPdfFileWriter(inf)
        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        with pytest.raises(SigningError, match='hybrid xrefs are disabled'):
            signers.sign_pdf(w, meta, signer=SELF_SIGN)


def test_allow_hybrid_sign():
    fname = 'minimal-hybrid-xref.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        w = IncrementalPdfFileWriter(inf, strict=False)
        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        signers.PdfSigner(signature_meta=meta, signer=SELF_SIGN).sign_pdf(w)


@freeze_time('2020-11-01')
def test_allow_hybrid_sign_validate_fail():
    fname = 'minimal-hybrid-xref.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        w = IncrementalPdfFileWriter(inf, strict=False)
        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        out = signers.PdfSigner(signature_meta=meta, signer=FROM_CA).sign_pdf(w)

    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    with pytest.raises(SignatureValidationError, match="do not permit.*hybrid"):
        val_trusted(s)


@freeze_time('2020-11-01')
def test_allow_hybrid_sign_validate_allow():
    fname = 'minimal-hybrid-xref.pdf'
    with open(os.path.join(PDF_DATA_DIR, fname), 'rb') as inf:
        w = IncrementalPdfFileWriter(inf, strict=False)
        meta = signers.PdfSignatureMetadata(field_name='Sig1')
        out = signers.PdfSigner(signature_meta=meta, signer=FROM_CA).sign_pdf(w)

    r = PdfFileReader(out, strict=False)
    s = r.embedded_signatures[0]

    vc = SIMPLE_V_CONTEXT()
    val_status = validate_pdf_signature(s, vc)
    assert val_status.bottom_line


@freeze_time('2020-11-01')
def test_ed25519():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1'),
        signer=FROM_ED25519_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'sha512'

    (extn,) = r.root['/Extensions']['/ISO_']
    assert extn['/ExtensionLevel'] == 32002


@freeze_time('2020-11-01')
def test_ed25519_trust():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1'),
        signer=FROM_ED25519_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s, vc=SIMPLE_ED25519_V_CONTEXT())


@freeze_time('2020-11-01')
def test_ed448():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_ED448_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'shake256_len'

    assert len(s.external_digest) == 64

    extn1, extn2 = r.root['/Extensions']['/ISO_']
    assert {extn1['/ExtensionLevel'], extn2['/ExtensionLevel']} == {
        32001,
        32002,
    }


@freeze_time('2020-11-01')
def test_ed448_trust():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w, signers.PdfSignatureMetadata(field_name='Sig1'), signer=FROM_ED448_CA
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s, vc=SIMPLE_ED448_V_CONTEXT())


@freeze_time('2020-11-01')
def test_ed448_invalid_hash_algo():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    with pytest.raises(SigningError, match='specifies.*shake256'):
        signers.sign_pdf(
            w,
            signers.PdfSignatureMetadata(
                field_name='Sig1', md_algorithm='sha256'
            ),
            signer=FROM_ED448_CA,
        )


@freeze_time('2020-11-01')
def test_ecdsa_with_sha256():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(field_name='Sig1', md_algorithm='sha256'),
        signer=FROM_ECC_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'sha256'

    (extn,) = r.root['/Extensions']['/ISO_']
    assert extn['/ExtensionLevel'] == 32002


@freeze_time('2020-11-01')
def test_ecdsa_with_sha3():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1', md_algorithm='sha3_256'
        ),
        signer=FROM_ECC_CA,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'sha3_256'

    extn1, extn2 = r.root['/Extensions']['/ISO_']
    assert {extn1['/ExtensionLevel'], extn2['/ExtensionLevel']} == {
        32001,
        32002,
    }


@freeze_time('2020-11-01')
def test_rsa_with_sha3():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        # need the generic mechanism because asn1crypto (==1.5.1)
        # doesn't have the OIDs for RSA-with-SHA3 family
        # hash functions.
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            md_algorithm='sha3_256',
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'sha3_256'

    (extn,) = r.root['/Extensions']['/ISO_']
    assert extn['/ExtensionLevel'] == 32001


@freeze_time('2020-11-01')
def test_rsa_with_sha384():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    signer = signers.SimpleSigner(
        signing_cert=FROM_CA.signing_cert,
        signing_key=FROM_CA.signing_key,
        cert_registry=FROM_CA.cert_registry,
        # need the generic mechanism because asn1crypto (==1.5.1)
        # doesn't have the OIDs for RSA-with-SHA3 family
        # hash functions.
        signature_mechanism=SignedDigestAlgorithm(
            {'algorithm': 'rsassa_pkcs1v15'}
        ),
    )
    out = signers.sign_pdf(
        w,
        signers.PdfSignatureMetadata(
            field_name='Sig1',
            md_algorithm='sha384',
        ),
        signer=signer,
    )
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    status = val_untrusted(s)
    assert status.md_algorithm == 'sha384'

    assert r.input_version == (1, 7)
    assert '/Extensions' not in r.root


@freeze_time('2020-11-01')
def test_sign_with_build_props_app_name():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', app_build_props=BuildProps(name='Test Application')
    )
    out = signers.sign_pdf(w, meta, signer=FROM_CA)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s)
    build_prop_dict = s.sig_object['/Prop_Build']['/App']
    assert build_prop_dict['/Name'] == '/Test Application'
    assert '/REx' not in build_prop_dict


@freeze_time('2020-11-01')
def test_sign_with_build_props_versioned_app_name():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        app_build_props=BuildProps(name='Test Application', revision='1.2.3'),
    )

    out = signers.sign_pdf(w, meta, signer=FROM_CA)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s)
    build_prop_dict = s.sig_object['/Prop_Build']['/App']
    assert build_prop_dict['/Name'] == '/Test Application'
    assert build_prop_dict['/REx'] == '1.2.3'


@freeze_time('2020-11-01')
def test_signature_dict_with_prop_auth_time():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    meta = signers.PdfSignatureMetadata(field_name='Sig1', prop_auth_time=512)

    out = signers.sign_pdf(w, meta, signer=FROM_CA)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s)
    assert s.sig_object['/Prop_AuthTime'] == 512


@freeze_time('2020-11-01')
def test_signature_dict_with_contact_info():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    contact_info = '+55 99 99999-9999'
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1', contact_info=contact_info
    )

    out = signers.sign_pdf(w, meta, signer=FROM_CA)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s)
    assert s.sig_object['/ContactInfo'] == contact_info


@freeze_time('2020-11-01')
def test_signature_dict_with_prop_auth_type():
    w = IncrementalPdfFileWriter(BytesIO(MINIMAL))
    auth_type = fields.SigAuthType.PASSWORD
    meta = signers.PdfSignatureMetadata(
        field_name='Sig1',
        prop_auth_type=auth_type,
    )

    out = signers.sign_pdf(w, meta, signer=FROM_CA)
    r = PdfFileReader(out)
    s = r.embedded_signatures[0]
    val_trusted(s)
    assert s.sig_object['/Prop_AuthType'] == 'Password'


def test_sign_reject_econtent_if_detached():
    fname = os.path.join(PDF_DATA_DIR, 'pdf-sig-with-econtent.pdf')
    with open(fname, 'rb') as inf:
        r = PdfFileReader(inf)
        emb = r.embedded_signatures[0]

        with pytest.raises(
            SignatureValidationError, match='detached.*encapsulated'
        ):
            val_untrusted(emb)



import logging
import os
from collections import namedtuple
from datetime import datetime
from typing import List, Optional, Union

from asn1crypto import cms, x509
from pyhanko.pdf_utils import generic, misc
from pyhanko.pdf_utils.generic import pdf_name
from pyhanko.pdf_utils.reader import PdfFileReader, process_data_at_eof
from pyhanko.sign.diff_analysis import (
    DEFAULT_DIFF_POLICY,
    DiffPolicy,
    DiffResult,
    ModificationLevel,
    SuspiciousModification,
)
from pyhanko.sign.fields import (
    FieldMDPSpec,
    MDPPerm,
    SeedLockDocument,
    SigSeedSubFilter,
    SigSeedValFlags,
    SigSeedValueSpec,
)
from pyhanko.sign.general import (
    SignedDataCerts,
    UnacceptableSignerError,
    byte_range_digest,
    extract_signer_info,
)

from pyhanko_certvalidator import ValidationContext
from pyhanko_certvalidator.path import ValidationPath

from .errors import (
    SignatureValidationError,
    SigSeedValueValidationError,
    ValidationInfoReadingError,
)
from .generic_cms import (
    cms_basic_validation,
    collect_signer_attr_status,
    collect_timing_info,
    compute_signature_tst_digest,
    extract_certs_for_validation,
    extract_self_reported_ts,
    extract_tst_data,
    validate_tst_signed_data,
)
from .settings import KeyUsageConstraints
from .status import (
    DocumentTimestampStatus,
    PdfSignatureStatus,
    SignatureCoverageLevel,
)
from .utils import CMSAlgorithmUsagePolicy

__all__ = [
    'EmbeddedPdfSignature',
    'DocMDPInfo',
    'read_certification_data',
    'async_validate_pdf_signature',
    'async_validate_pdf_timestamp',
    'report_seed_value_validation',
    'extract_contents',
]


logger = logging.getLogger(__name__)


def _extract_reference_dict(
    signature_obj, method
) -> Optional[generic.DictionaryObject]:
    try:
        sig_refs = signature_obj['/Reference']
    except KeyError:
        return None
    for ref in sig_refs:
        ref = ref.get_object()
        if ref['/TransformMethod'] == method:
            return ref
    return None


def _extract_docmdp_for_sig(signature_obj) -> Optional[MDPPerm]:
    ref = _extract_reference_dict(signature_obj, '/DocMDP')
    if ref is None:
        return None
    try:
        raw_perms = ref['/TransformParams'].raw_get('/P')
        return MDPPerm(raw_perms)
    except (ValueError, KeyError) as e:  # pragma: nocover
        raise SignatureValidationError(
            "Failed to read document permissions"
        ) from e


def extract_contents(sig_object: generic.DictionaryObject) -> bytes:
    """
    Internal function to extract the (DER-encoded) signature bytes from a PDF
    signature dictionary.

    :param sig_object:
        A signature dictionary.
    :return:
        The extracted contents as a byte string.
    """

    try:
        cms_content = sig_object.raw_get(
            '/Contents', decrypt=generic.EncryptedObjAccess.RAW
        )
    except KeyError:
        raise misc.PdfReadError('Could not read /Contents entry in signature')

    if not isinstance(
        cms_content, (generic.TextStringObject, generic.ByteStringObject)
    ):
        raise misc.PdfReadError('/Contents must be string-like')
    return cms_content.original_bytes


# TODO clarify in docs that "external timestamp" is always None when dealing
#  with a /DocTimeStamp, since there the timestamp token is simply the entire
#  signature object
class EmbeddedPdfSignature:
    """
    Class modelling a signature embedded in a PDF document.
    """

    sig_field: generic.DictionaryObject
    """
    The field dictionary of the form field containing the signature.
    """

    sig_object: generic.DictionaryObject
    """
    The signature dictionary.
    """

    signed_data: cms.SignedData
    """
    CMS signed data in the signature.
    """

    def __init__(
        self,
        reader: PdfFileReader,
        sig_field: generic.DictionaryObject,
        fq_name: str,
    ):
        self.reader = reader
        if isinstance(sig_field, generic.IndirectObject):
            sig_field = sig_field.get_object()
        self.sig_field = sig_field
        sig_object_ref = sig_field.raw_get('/V')
        self.sig_object = sig_object = sig_object_ref.get_object()
        assert isinstance(sig_object, generic.DictionaryObject)
        try:
            self.byte_range = sig_object.raw_get('/ByteRange')
        except KeyError:
            raise misc.PdfReadError(
                'Could not read /ByteRange entry in signature'
            )
        self.pkcs7_content = cms_content = extract_contents(sig_object)

        message = cms.ContentInfo.load(cms_content)
        signed_data = message['content']
        self.signed_data: cms.SignedData = signed_data

        self.signer_info = extract_signer_info(signed_data)
        self._sd_cert_info: Optional[SignedDataCerts] = None

        # The PDF standard does not define a way to specify the digest algorithm
        # used other than this one.
        # However, RFC 5652 § 11.2 states that the message_digest attribute
        # (which in our case is the PDF's ByteRange digest) is to be computed
        # using the signer's digest algorithm. This can only refer
        # to the corresponding SignerInfo entry.
        digest_algo = self.signer_info['digest_algorithm']
        self.md_algorithm = digest_algo['algorithm'].native.lower()
        eci = signed_data['encap_content_info']
        content_type = eci['content_type'].native
        if content_type == 'data':
            # Case of a normal signature
            self.external_md_algorithm = self.md_algorithm
        elif content_type == 'tst_info':
            # for timestamps, the hash algorithm in the messageImprint
            # need not be the same as the one to digest the encapsulated data!
            # RFC 8933 recommends to unify them, but it's not a given.
            mi = eci['content'].parsed['message_imprint']
            self.external_md_algorithm = mi['hash_algorithm'][
                'algorithm'
            ].native

        # grab the revision to which the signature applies
        # NOTE: We're using get_last_change here as opposed to
        # get_introducing_revision. The distinction won't be relevant in most
        # legitimate use cases, but get_last_change is more likely to be correct
        # in cases where the signature obj was created by overriding an existing
        # object (which is weird, but technically possible, I guess).
        # Important note: the coverage checker will validate whether the
        # xref table for that revision is actually covered by the signature,
        # and raise the alarm if that's not the case.
        # Therefore shenanigans with updating signature objects will be detected
        # even before the diff checker runs.
        self.signed_revision = self.reader.xrefs.get_last_change(
            sig_object_ref.reference
        )
        self.coverage = None
        self.external_digest: Optional[bytes] = None
        self.total_len: Optional[int] = None
        self._docmdp: Optional[MDPPerm] = None
        self._fieldmdp: Optional[FieldMDPSpec] = None
        self._docmdp_queried = self._fieldmdp_queried = False
        self.tst_signature_digest: Optional[bytes] = None

        self.diff_result = None
        self._integrity_checked = False
        self.fq_name = fq_name

    def _init_cert_info(self) -> SignedDataCerts:
        if self._sd_cert_info is None:
            self._sd_cert_info = extract_certs_for_validation(self.signed_data)
        return self._sd_cert_info

    @property
    def embedded_attr_certs(self) -> List[cms.AttributeCertificateV2]:
        """
        Embedded attribute certificates.
        """
        return list(self._init_cert_info().attribute_certs)

    @property
    def other_embedded_certs(self) -> List[x509.Certificate]:
        """
        Embedded X.509 certificates, excluding than that of the signer.
        """
        return list(self._init_cert_info().other_certs)

    @property
    def signer_cert(self) -> x509.Certificate:
        """
        Certificate of the signer.
        """
        return self._init_cert_info().signer_cert

    @property
    def sig_object_type(self) -> generic.NameObject:
        """
        Returns the type of the embedded signature object.
        For ordinary signatures, this will be ``/Sig``.
        In the case of a document timestamp, ``/DocTimeStamp`` is returned.

        :return:
            A PDF name object describing the type of signature.
        """
        return self.sig_object.get('/Type', pdf_name('/Sig'))

    @property
    def field_name(self):
        """
        :return:
            Name of the signature field.
        """
        return self.fq_name

    @property
    def self_reported_timestamp(self) -> Optional[datetime]:
        """
        :return:
            The signing time as reported by the signer, if embedded in the
            signature's signed attributes or provided as part of the signature
            object in the PDF document.
        """
        ts = extract_self_reported_ts(self.signer_info)
        if ts is not None:
            return ts

        try:
            st_as_pdf_date = self.sig_object['/M']
            return generic.parse_pdf_date(
                st_as_pdf_date, strict=self.reader.strict
            )
        except KeyError:  # pragma: nocover
            return None

    @property
    def attached_timestamp_data(self) -> Optional[cms.SignedData]:
        """
        :return:
            The signed data component of the timestamp token embedded in this
            signature, if present.
        """
        return extract_tst_data(self.signer_info)

    def compute_integrity_info(self, diff_policy=None, skip_diff=False):
        """
        Compute the various integrity indicators of this signature.

        :param diff_policy:
            Policy to evaluate potential incremental updates that were appended
            to the signed revision of the document.
            Defaults to
            :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
        :param skip_diff:
            If ``True``, skip the difference analysis step entirely.
        """
        self._enforce_hybrid_xref_policy()
        self.compute_digest()
        self.compute_tst_digest()

        # TODO in scenarios where we have to verify multiple signatures, we're
        #  doing a lot of double work here. This could be improved.
        self.coverage = self.evaluate_signature_coverage()
        diff_policy = diff_policy or DEFAULT_DIFF_POLICY
        if not skip_diff:
            self.diff_result = self.evaluate_modifications(diff_policy)

        self._integrity_checked = True

    def summarise_integrity_info(self) -> dict:
        """
        Compile the integrity information for this signature into a dictionary
        that can later be passed to :class:`.PdfSignatureStatus` as kwargs.

        This method is only available after calling
        :meth:`.EmbeddedPdfSignature.compute_integrity_info`.
        """

        if not self._integrity_checked:
            raise SignatureValidationError(
                "Call compute_integrity_info() before invoking"
                "summarise_integrity_info()"
            )  # pragma: nocover

        docmdp = self.docmdp_level
        diff_result = self.diff_result
        coverage = self.coverage
        docmdp_ok = None

        # attempt to set docmdp_ok based on the diff analysis results
        if diff_result is not None:
            mod_level = (
                diff_result.modification_level
                if isinstance(diff_result, DiffResult)
                else ModificationLevel.OTHER
            )
            docmdp_ok = not (
                mod_level == ModificationLevel.OTHER
                or (docmdp is not None and mod_level.value > docmdp.value)
            )
        elif coverage != SignatureCoverageLevel.ENTIRE_REVISION:
            # if the diff analysis didn't run, we can still do something
            # meaningful if coverage is not ENTIRE_REVISION:
            #  - if the signature covers the entire file, we're good.
            #  - if the coverage level is anything else, not so much
            docmdp_ok = coverage == SignatureCoverageLevel.ENTIRE_FILE

        status_kwargs = {
            'coverage': coverage,
            'docmdp_ok': docmdp_ok,
            'diff_result': diff_result,
        }
        return status_kwargs

    @property
    def seed_value_spec(self) -> Optional[SigSeedValueSpec]:
        try:
            sig_sv_dict = self.sig_field['/SV']
        except KeyError:
            return None
        return SigSeedValueSpec.from_pdf_object(sig_sv_dict)

    @property
    def docmdp_level(self) -> Optional[MDPPerm]:
        """
        :return:
            The document modification policy required by this signature or
            its Lock dictionary.

            .. warning::
                This does not take into account the DocMDP requirements of
                earlier signatures (if present).

                The specification forbids signing with a more lenient DocMDP
                than the one currently in force, so this should not happen
                in a compliant document.
                That being said, any potential violations will still invalidate
                the earlier signature with the stricter DocMDP policy.

        """
        if self._docmdp_queried:
            return self._docmdp
        docmdp = _extract_docmdp_for_sig(signature_obj=self.sig_object)

        if docmdp is None:
            try:
                lock_dict = self.sig_field['/Lock']
                docmdp = MDPPerm(lock_dict['/P'])
            except KeyError:
                pass
        self._docmdp = docmdp
        self._docmdp_queried = True
        return docmdp

    @property
    def fieldmdp(self) -> Optional[FieldMDPSpec]:
        """
        :return:
            Read the field locking policy of this signature, if applicable.
            See also :class:`~.pyhanko.sign.fields.FieldMDPSpec`.
        """
        # TODO as above, fall back to /Lock
        if self._fieldmdp_queried:
            return self._fieldmdp
        ref_dict = _extract_reference_dict(self.sig_object, '/FieldMDP')
        self._fieldmdp_queried = True
        if ref_dict is None:
            return None
        try:
            sp = FieldMDPSpec.from_pdf_object(ref_dict['/TransformParams'])
        except (ValueError, KeyError) as e:  # pragma: nocover
            raise SignatureValidationError(
                "Failed to read /FieldMDP settings"
            ) from e
        self._fieldmdp = sp
        return sp

    def compute_digest(self) -> bytes:
        """
        Compute the ``/ByteRange`` digest of this signature.
        The result will be cached.

        :return:
            The digest value.
        """
        if self.external_digest is not None:
            return self.external_digest

        self.total_len, digest = byte_range_digest(
            self.reader.stream,
            byte_range=self.byte_range,
            md_algorithm=self.external_md_algorithm,
        )
        self.external_digest = digest
        return digest

    def compute_tst_digest(self) -> Optional[bytes]:
        """
        Compute the digest of the signature needed to validate its timestamp
        token (if present).

        .. warning::
            This computation is only relevant for timestamp tokens embedded
            inside a regular signature.
            If the signature in question is a document timestamp (where the
            entire signature object is a timestamp token), this method
            does not apply.

        :return:
            The digest value, or ``None`` if there is no timestamp token.
        """

        if self.tst_signature_digest is not None:
            return self.tst_signature_digest
        self.tst_signature_digest = digest = compute_signature_tst_digest(
            self.signer_info
        )
        return digest

    def evaluate_signature_coverage(self) -> SignatureCoverageLevel:
        """
        Internal method used to evaluate the coverage level of a signature.

        :return:
            The coverage level of the signature.
        """

        xref_cache = self.reader.xrefs
        # for the coverage check, we're more strict with regards to the byte
        #  range
        stream = self.reader.stream

        # nonstandard byte range -> insta-fail
        if len(self.byte_range) != 4 or self.byte_range[0] != 0:
            return SignatureCoverageLevel.UNCLEAR

        _, len1, start2, len2 = self.byte_range

        # first check: check if the signature covers the entire file.
        #  (from a cryptographic point of view)
        # In this case, there are no changes at all, so we're good.

        # compute file size
        stream.seek(0, os.SEEK_END)
        # the * 2 is because of the ASCII hex encoding, and the + 2
        # is the wrapping <>
        embedded_sig_content = len(self.pkcs7_content) * 2 + 2
        signed_zone_len = len1 + len2 + embedded_sig_content
        file_covered = stream.tell() == signed_zone_len
        if file_covered:
            return SignatureCoverageLevel.ENTIRE_FILE

        # Now we're in the mixed case: the byte range is a standard one
        #  starting at the beginning of the document, but it doesn't go all
        #  the way to the end of the file. This can be for legitimate reasons,
        #  not all of which we can evaluate right now.

        # First, check if the signature is a contiguous one.
        # In other words, we need to check if the interruption in the byte
        # range is "fully explained" by the signature content.
        contiguous = start2 == len1 + embedded_sig_content
        if not contiguous:
            return SignatureCoverageLevel.UNCLEAR

        # next, we verify that the revision this signature belongs to
        #  is completely covered. This requires a few separate checks.
        # (1) Verify that the xref container (table or stream) is covered
        # (2) Verify the presence of the EOF and startxref markers at the
        #     end of the signed region, and compare them with the values
        #     in the xref cache to make sure we are reading the right revision.

        # Check (2) first, since it's the quickest
        stream.seek(signed_zone_len)
        signed_rev = self.signed_revision
        try:
            startxref = process_data_at_eof(stream)
            expected = xref_cache.get_startxref_for_revision(signed_rev)
            if startxref != expected:
                return SignatureCoverageLevel.CONTIGUOUS_BLOCK_FROM_START
        except misc.PdfReadError:
            return SignatureCoverageLevel.CONTIGUOUS_BLOCK_FROM_START

        # ... then check (1) for all revisions up to and including
        # signed_revision
        for revision in range(signed_rev + 1):
            xref_meta = xref_cache.get_xref_container_info(revision)
            if xref_meta.end_location > signed_zone_len:
                return SignatureCoverageLevel.CONTIGUOUS_BLOCK_FROM_START

        return SignatureCoverageLevel.ENTIRE_REVISION

    def _enforce_hybrid_xref_policy(self):
        reader = self.reader
        if reader.strict and reader.xrefs.hybrid_xrefs_present:
            raise SignatureValidationError(
                "Settings do not permit validation of signatures in "
                "hybrid-reference files."
            )

    def evaluate_modifications(
        self, diff_policy: DiffPolicy
    ) -> Union[DiffResult, SuspiciousModification]:
        """
        Internal method used to evaluate the modification level of a signature.
        """

        if self.coverage < SignatureCoverageLevel.ENTIRE_REVISION:
            return SuspiciousModification(
                'Nonstandard signature coverage level'
            )
        elif self.coverage == SignatureCoverageLevel.ENTIRE_FILE:
            return DiffResult(ModificationLevel.NONE, set())

        return diff_policy.review_file(
            self.reader,
            self.signed_revision,
            field_mdp_spec=self.fieldmdp,
            doc_mdp=self.docmdp_level,
        )


DocMDPInfo = namedtuple('DocMDPInfo', ['permission', 'author_sig'])
"""
Encodes certification information for a signed document, consisting of a
reference to the author signature, together with the associated DocMDP policy.
"""


def read_certification_data(reader: PdfFileReader) -> Optional[DocMDPInfo]:
    """
    Read the certification information for a PDF document, if present.

    :param reader:
        Reader representing the input document.
    :return:
        A :class:`.DocMDPInfo` object containing the relevant data, or ``None``.
    """
    try:
        certification_sig = reader.root['/Perms']['/DocMDP']
    except KeyError:
        return None

    perm = _extract_docmdp_for_sig(certification_sig)
    return DocMDPInfo(perm, certification_sig)


def _validate_sv_constraints(
    emb_sig: EmbeddedPdfSignature, validation_path, timestamp_found
):
    sv_spec = emb_sig.seed_value_spec
    if sv_spec is None:
        return
    signing_cert = emb_sig.signer_cert
    if sv_spec.cert is not None:
        try:
            sv_spec.cert.satisfied_by(signing_cert, validation_path)
        except UnacceptableSignerError as e:
            raise SigSeedValueValidationError(e) from e

    if not timestamp_found and sv_spec.timestamp_required:
        raise SigSeedValueValidationError(
            "The seed value dictionary requires a trusted timestamp, but "
            "none was found, or the timestamp did not validate."
        )

    sig_obj = emb_sig.sig_object

    if sv_spec.seed_signature_type is not None:
        sv_certify = sv_spec.seed_signature_type.certification_signature()
        try:
            perms: generic.DictionaryObject = emb_sig.reader.root['/Perms']
            cert_sig_ref = perms.get_value_as_reference('/DocMDP')
            was_certified = cert_sig_ref == sig_obj.container_ref
        except (KeyError, generic.IndirectObjectExpected, AttributeError):
            was_certified = False
        if sv_certify != was_certified:

            def _type(certify):
                return 'a certification' if certify else 'an approval'

            raise SigSeedValueValidationError(
                "The seed value dictionary's /MDP entry specifies that "
                f"this field should contain {_type(sv_certify)} "
                f"signature, but {_type(was_certified)} "
                "appears to have been used."
            )
        if sv_certify:
            sv_mdp_perm = sv_spec.seed_signature_type.mdp_perm
            doc_mdp = emb_sig.docmdp_level
            if sv_mdp_perm != doc_mdp:
                raise SigSeedValueValidationError(
                    "The seed value dictionary specified that this "
                    "certification signature should use the MDP policy "
                    f"'{sv_mdp_perm}', but '{doc_mdp}' was "
                    "used in the signature."
                )

    flags = sv_spec.flags
    if not flags:
        return

    selected_sf_str = sig_obj['/SubFilter']
    selected_sf = SigSeedSubFilter(selected_sf_str)
    if (flags & SigSeedValFlags.SUBFILTER) and sv_spec.subfilters is not None:
        # empty array = no supported subfilters
        if not sv_spec.subfilters:
            raise NotImplementedError(
                "The signature encodings mandated by the seed value "
                "dictionary are not supported."
            )
        # standard mandates that we take the first available subfilter
        mandated_sf: SigSeedSubFilter = sv_spec.subfilters[0]
        if selected_sf is not None and mandated_sf != selected_sf:
            raise SigSeedValueValidationError(
                "The seed value dictionary mandates subfilter '%s', "
                "but '%s' was used in the signature."
                % (mandated_sf.value, selected_sf.value)
            )

    if (
        flags & SigSeedValFlags.APPEARANCE_FILTER
    ) and sv_spec.appearance is not None:
        logger.warning(
            "The signature's seed value dictionary specifies the "
            "/AppearanceFilter entry as mandatory, but this constraint "
            "is impossible to validate."
        )

    if (
        flags & SigSeedValFlags.LEGAL_ATTESTATION
    ) and sv_spec.legal_attestations is not None:
        raise NotImplementedError(
            "pyHanko does not support legal attestations, but the seed value "
            "dictionary mandates that they be restricted to a specific subset."
        )

    if (
        flags & SigSeedValFlags.LOCK_DOCUMENT
    ) and sv_spec.lock_document is not None:
        doc_mdp = emb_sig.docmdp_level
        if (
            sv_spec.lock_document == SeedLockDocument.LOCK
            and doc_mdp != MDPPerm.NO_CHANGES
        ):
            raise SigSeedValueValidationError(
                "Document must be locked, but some changes are still allowed."
            )
        if (
            sv_spec.lock_document == SeedLockDocument.DO_NOT_LOCK
            and doc_mdp == MDPPerm.NO_CHANGES
        ):
            raise SigSeedValueValidationError(
                "Document must not be locked, but the DocMDP level is set to "
                "NO_CHANGES."
            )
        # value 'auto' is OK.

    signer_info = emb_sig.signer_info
    if (
        flags & SigSeedValFlags.ADD_REV_INFO
    ) and sv_spec.add_rev_info is not None:
        from pyhanko.sign.validation.ltv import retrieve_adobe_revocation_info

        try:
            retrieve_adobe_revocation_info(signer_info)
            revinfo_found = True
        except ValidationInfoReadingError:
            revinfo_found = False

        if sv_spec.add_rev_info != revinfo_found:
            raise SigSeedValueValidationError(
                "The seed value dict mandates that revocation info %sbe "
                "added, but it was %sfound in the signature."
                % (
                    "" if sv_spec.add_rev_info else "not ",
                    "" if revinfo_found else "not ",
                )
            )
        if (
            sv_spec.add_rev_info
            and selected_sf != SigSeedSubFilter.ADOBE_PKCS7_DETACHED
        ):
            raise SigSeedValueValidationError(
                "The seed value dict mandates that Adobe-style revocation "
                "info be added; this requires subfilter '%s'"
                % (SigSeedSubFilter.ADOBE_PKCS7_DETACHED.value)
            )

    if (
        flags & SigSeedValFlags.DIGEST_METHOD
    ) and sv_spec.digest_methods is not None:
        selected_md = emb_sig.md_algorithm.lower()
        if selected_md not in sv_spec.digest_methods:
            raise SigSeedValueValidationError(
                "The selected message digest %s is not allowed by the "
                "seed value dictionary." % selected_md
            )

    if flags & SigSeedValFlags.REASONS:
        # standard says that omission of the /Reasons key amounts to
        #  a prohibition in this case
        reasons = sv_spec.reasons or []
        must_omit = not reasons or reasons == ["."]
        reason_given = sig_obj.get('/Reason')
        if must_omit and reason_given is not None:
            raise SigSeedValueValidationError(
                "The seed value dictionary prohibits giving a reason "
                "for signing."
            )
        if not must_omit and reason_given not in reasons:
            raise SigSeedValueValidationError(
                "The reason for signing \"%s\" is not accepted by the "
                "seed value dictionary." % (reason_given,)
            )


def report_seed_value_validation(
    embedded_sig: EmbeddedPdfSignature,
    validation_path: ValidationPath,
    timestamp_found: bool,
):
    """
    Internal API function to enforce seed value constraints (if present)
    and report on the result(s).

    :param embedded_sig:
        The embedded signature.
    :param validation_path:
        The validation path for the signer's certificate.
    :param timestamp_found:
        Flag indicating whether a valid timestamp was found or not.
    :return:
        A ``status_kwargs`` dict.
    """
    sv_err: Optional[SigSeedValueValidationError]
    try:
        _validate_sv_constraints(
            embedded_sig, validation_path, timestamp_found=timestamp_found
        )
        sv_err = None
    except SigSeedValueValidationError as e:
        logger.warning("Error in seed value validation.", exc_info=e)
        sv_err = e
    return {
        'has_seed_values': embedded_sig.seed_value_spec is not None,
        'seed_value_constraint_error': sv_err,
    }


def _validate_subfilter(subfilter_str, permitted_subfilters, err_msg):
    try:
        from pyhanko.sign.fields import SigSeedSubFilter

        subfilter_ok = SigSeedSubFilter(subfilter_str) in permitted_subfilters
    except ValueError:
        subfilter_ok = False

    if not subfilter_ok:
        raise SignatureValidationError(err_msg % subfilter_str)


async def async_validate_pdf_signature(
    embedded_sig: EmbeddedPdfSignature,
    signer_validation_context: Optional[ValidationContext] = None,
    ts_validation_context: Optional[ValidationContext] = None,
    ac_validation_context: Optional[ValidationContext] = None,
    diff_policy: Optional[DiffPolicy] = None,
    key_usage_settings: Optional[KeyUsageConstraints] = None,
    skip_diff: bool = False,
    algorithm_policy: Optional[CMSAlgorithmUsagePolicy] = None,
) -> PdfSignatureStatus:
    """
    .. versionadded:: 0.9.0

    .. versionchanged: 0.11.0
        Added ``ac_validation_context`` param.


    Validate a PDF signature.

    :param embedded_sig:
        Embedded signature to evaluate.
    :param signer_validation_context:
        Validation context to use to validate the signature's chain of trust.
    :param ts_validation_context:
        Validation context to use to validate the timestamp's chain of trust
        (defaults to ``signer_validation_context``).
    :param ac_validation_context:
        Validation context to use to validate attribute certificates.
        If not supplied, no AC validation will be performed.

        .. note::
            :rfc:`5755` requires attribute authority trust roots to be specified
            explicitly; hence why there's no default.
    :param diff_policy:
        Policy to evaluate potential incremental updates that were appended
        to the signed revision of the document.
        Defaults to
        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
    :param key_usage_settings:
        A :class:`.KeyUsageConstraints` object specifying which key usages
        must or must not be present in the signer's certificate.
    :param skip_diff:
        If ``True``, skip the difference analysis step entirely.
    :param algorithm_policy:
        The algorithm usage policy for the signature validation.

        .. warning::
            This is distinct from the algorithm usage policy used for
            certificate validation, but the latter will be used as a fallback
            if this parameter is not specified.

            It is nonetheless recommended to align both policies unless
            there is a clear reason to do otherwise.
    :return:
        The status of the PDF signature in question.
    """

    sig_object = embedded_sig.sig_object
    if embedded_sig.sig_object_type != '/Sig':
        raise SignatureValidationError("Signature object type must be /Sig")

    # check whether the subfilter type is one we support
    subfilter_str = sig_object.get('/SubFilter', None)
    _validate_subfilter(
        subfilter_str,
        (SigSeedSubFilter.ADOBE_PKCS7_DETACHED, SigSeedSubFilter.PADES),
        "%s is not a recognized SubFilter type in signatures.",
    )

    if ts_validation_context is None:
        ts_validation_context = signer_validation_context

    embedded_sig.compute_integrity_info(
        diff_policy=diff_policy, skip_diff=skip_diff
    )
    status_kwargs = embedded_sig.summarise_integrity_info()

    ts_status_kwargs = await collect_timing_info(
        embedded_sig.signer_info,
        ts_validation_context,
        raw_digest=embedded_sig.compute_digest(),
    )
    status_kwargs.update(ts_status_kwargs)
    if 'signer_reported_dt' not in status_kwargs:
        # maybe the PDF signature dictionary declares /M
        signer_reported_dt = embedded_sig.self_reported_timestamp
        if signer_reported_dt is not None:
            status_kwargs['signer_reported_dt'] = signer_reported_dt

    key_usage_settings = PdfSignatureStatus.default_usage_constraints(
        key_usage_settings
    )
    status_kwargs = await cms_basic_validation(
        embedded_sig.signed_data,
        raw_digest=embedded_sig.external_digest,
        validation_context=signer_validation_context,
        status_kwargs=status_kwargs,
        key_usage_settings=key_usage_settings,
        algorithm_policy=algorithm_policy,
    )
    tst_validity = status_kwargs.get('timestamp_validity', None)
    timestamp_found = (
        tst_validity is not None and tst_validity.valid and tst_validity.trusted
    )
    sv_update = report_seed_value_validation(
        embedded_sig, status_kwargs['validation_path'], timestamp_found
    )
    status_kwargs.update(sv_update)
    if ac_validation_context is not None:
        ac_validation_context.certificate_registry.register_multiple(
            embedded_sig.other_embedded_certs
        )
    status_kwargs.update(
        await collect_signer_attr_status(
            sd_attr_certificates=embedded_sig.embedded_attr_certs,
            signer_cert=embedded_sig.signer_cert,
            validation_context=ac_validation_context,
            sd_signed_attrs=embedded_sig.signer_info['signed_attrs'],
        )
    )
    return PdfSignatureStatus(**status_kwargs)


async def async_validate_pdf_timestamp(
    embedded_sig: EmbeddedPdfSignature,
    validation_context: Optional[ValidationContext] = None,
    diff_policy: Optional[DiffPolicy] = None,
    skip_diff: bool = False,
) -> DocumentTimestampStatus:
    """
    .. versionadded:: 0.9.0

    Validate a PDF document timestamp.

    :param embedded_sig:
        Embedded signature to evaluate.
    :param validation_context:
        Validation context to use to validate the timestamp's chain of trust.
    :param diff_policy:
        Policy to evaluate potential incremental updates that were appended
        to the signed revision of the document.
        Defaults to
        :const:`~pyhanko.sign.diff_analysis.DEFAULT_DIFF_POLICY`.
    :param skip_diff:
        If ``True``, skip the difference analysis step entirely.
    :return:
        The status of the PDF timestamp in question.
    """

    if embedded_sig.sig_object_type != '/DocTimeStamp':
        raise SignatureValidationError(
            "Signature object type must be /DocTimeStamp"
        )

    # check whether the subfilter type is one we support
    subfilter_str = embedded_sig.sig_object.get('/SubFilter', None)
    _validate_subfilter(
        subfilter_str,
        (SigSeedSubFilter.ETSI_RFC3161,),
        "%s is not a recognized SubFilter type for timestamps.",
    )

    embedded_sig.compute_integrity_info(
        diff_policy=diff_policy, skip_diff=skip_diff
    )

    status_kwargs = await validate_tst_signed_data(
        embedded_sig.signed_data,
        validation_context,
        embedded_sig.compute_digest(),
    )

    status_kwargs['coverage'] = embedded_sig.coverage
    status_kwargs['diff_result'] = embedded_sig.diff_result
    return DocumentTimestampStatus(**status_kwargs)



import requests
from asn1crypto import tsp

from pyhanko_certvalidator._asyncio_compat import to_thread

from .api import TimeStamper
from .common_utils import TimestampRequestError, set_tsp_headers

__all__ = ['HTTPTimeStamper']


class HTTPTimeStamper(TimeStamper):
    """
    Standard HTTP-based timestamp client.
    """

    def __init__(self, url, https=False, timeout=5, auth=None, headers=None):
        """
        Initialise the timestamp client.

        :param url:
            URL where the server listens for timestamp requests.
        :param https:
            Enforce HTTPS.
        :param timeout:
            Timeout (in seconds)
        :param auth:
            Value of HTTP ``Authorization`` header
        :param headers:
            Other headers to include.
        """
        if https and not url.startswith('https:'):  # pragma: nocover
            raise ValueError('Timestamp URL is not HTTPS.')
        self.url = url
        self.timeout = timeout
        self.auth = auth
        self.headers = headers
        super().__init__()

    def request_headers(self) -> dict:
        """
        Format the HTTP request headers.

        :return:
            Header dictionary.
        """
        return set_tsp_headers(self.headers or {})

    async def async_request_tsa_response(
        self, req: tsp.TimeStampReq
    ) -> tsp.TimeStampResp:
        def task():
            try:
                raw_res = requests.post(
                    self.url,
                    req.dump(),
                    headers=self.request_headers(),
                    auth=self.auth,
                    timeout=self.timeout,
                )
            except IOError as e:
                raise TimestampRequestError(
                    'Error in communication with timestamp server',
                ) from e

            if (
                raw_res.headers.get('Content-Type')
                != 'application/timestamp-reply'
            ):
                raise TimestampRequestError(
                    'Timestamp server response is malformed.', raw_res
                )
            return tsp.TimeStampResp.load(raw_res.content)

        response = await to_thread(task)
        return response



from typing import Iterable, Tuple

from pyhanko.pdf_utils.reader import HistoricalResolver, RawPdfPath

from ..commons import compare_dicts, compare_key_refs, qualify_transforming
from ..constants import ROOT_EXEMPT_STRICT_COMPARISON
from ..policy_api import ModificationLevel
from ..rules_api import (
    Context,
    QualifiedWhitelistRule,
    ReferenceUpdate,
    WhitelistRule,
)

__all__ = ['CatalogModificationRule', 'ObjectStreamRule', 'XrefStreamRule']


class CatalogModificationRule(QualifiedWhitelistRule):
    """
    Rule that adjudicates modifications to the document catalog.

    :param ignored_keys:
        Values in the document catalog that may change between revisions.
        The default ones are ``/AcroForm``, ``/DSS``, ``/Extensions``,
        ``/Metadata``, ``/MarkInfo`` and ``/Version``.

        Checking for ``/AcroForm``, ``/DSS`` and ``/Metadata`` is delegated to
        :class:`.FormUpdatingRule`, :class:`.DSSCompareRule` and
        :class:`.MetadataUpdateRule`, respectively.
    """

    def __init__(self, ignored_keys=None):
        self.ignored_keys = (
            ignored_keys
            if ignored_keys is not None
            else ROOT_EXEMPT_STRICT_COMPARISON
        )

    def apply_qualified(
        self, old: HistoricalResolver, new: HistoricalResolver
    ) -> Iterable[Tuple[ModificationLevel, ReferenceUpdate]]:
        old_root = old.root
        new_root = new.root
        # first, check if the keys in the document catalog are unchanged
        compare_dicts(old_root, new_root, self.ignored_keys)

        # As for the keys in the root dictionary that are allowed to change:
        #  - /Extensions requires no further processing since it must consist
        #    of direct objects anyway, BUT the /Extensions entry itself can
        #    be indirect
        #  - /MarkInfo: if it's an indirect reference (probably not) we can
        #    whitelist it (but do not process any further)
        #  - /DSS, /AcroForm and /Metadata are dealt with by other rules.
        for key in ('/Extensions', '/MarkInfo'):
            yield from qualify_transforming(
                ModificationLevel.LTA_UPDATES,
                compare_key_refs(key, old, old_root, new_root),
                transform=lambda ref: ReferenceUpdate(
                    ref,
                    context_checked=Context.from_absolute(
                        old, RawPdfPath('/Root', key)
                    ),
                ),
            )

        yield ModificationLevel.LTA_UPDATES, ReferenceUpdate(
            new.root_ref,
            # Things like /Data in a MDP policy can point to root
            # and since we checked with compare_dicts, doing a blanket
            # approval is much easier than figuring out all the ways
            # in which /Root can be cross-referenced.
            context_checked=None,
        )


class ObjectStreamRule(WhitelistRule):
    """
    Rule that allows object streams to be added.

    Note that this rule only whitelists the object streams themselves (provided
    they do not override any existing objects, obviously), not the objects
    in them.
    """

    def apply(
        self, old: HistoricalResolver, new: HistoricalResolver
    ) -> Iterable[ReferenceUpdate]:
        # object streams are OK, but overriding object streams is not.
        for objstream_ref in new.object_streams_used():
            if old.is_ref_available(objstream_ref):
                yield ReferenceUpdate(objstream_ref)


class XrefStreamRule(WhitelistRule):
    """
    Rule that allows new cross-reference streams to be defined.
    """

    def apply(
        self, old: HistoricalResolver, new: HistoricalResolver
    ) -> Iterable[ReferenceUpdate]:
        xrefs = new.reader.xrefs
        xref_meta = xrefs.get_xref_container_info(new.revision)
        xref_stm = xref_meta.stream_ref
        if xref_stm is not None and old.is_ref_available(xref_stm):
            yield ReferenceUpdate(xref_stm)

        # If this revision is followed by a hybrid one, then we must
        #  clear the ref to the hybrid stream as well. Let's take care of that.

        # Note: this check is only relevant in nonstrict mode because
        #  hybrid-reference docs are banned otherwise.
        if new.reader.strict:
            return

        try:
            next_rev_data = xrefs.get_xref_data(new.revision + 1)
        except IndexError:
            return

        if next_rev_data.hybrid is not None:
            hyb_xref_stm = next_rev_data.hybrid.meta_info.stream_ref
            if hyb_xref_stm and old.is_ref_available(hyb_xref_stm):
                yield ReferenceUpdate(hyb_xref_stm)



from dataclasses import dataclass, field
from typing import Optional

from pyhanko.cli.config import CLIConfig
from pyhanko.sign import PdfSignatureMetadata
from pyhanko.sign.fields import SigFieldSpec
from pyhanko.stamp import BaseStampStyle


@dataclass
class UXContext:
    """
    Context object to track information that affects the UX, e.g. user intent
    as inferred from certain argument combinations that are otherwise difficult
    to wire throughout the UI code.
    """

    visible_signature_desired: bool = False
    """
    Set to `True` if the user explicitly specifies `--field` with a bounding box
    or passes `--style-name` explicitly.
    """


@dataclass
class CLIContext:
    """
    Context object that cobbles together various CLI settings values that were
    gathered by various subcommands during the lifetime of a CLI invocation,
    either from configuration or from command line arguments.
    This object is passed around as a ``click`` context object.

    Not all settings are applicable to all subcommands, so all values are
    optional.
    """

    sig_settings: Optional[PdfSignatureMetadata] = None
    """
    The settings that will be used to produce a new signature.
    """

    config: Optional[CLIConfig] = None
    """
    Values for CLI configuration settings.
    """

    existing_fields_only: bool = False
    """
    Whether signing operations should use existing fields only.
    """

    timestamp_url: Optional[str] = None
    """
    Endpoint URL for the timestamping service to use.
    """

    stamp_style: Optional[BaseStampStyle] = None
    """
    Stamp style to use for generating visual signature appearances, if
    applicable.
    """

    stamp_url: Optional[str] = None
    """
    For QR stamp styles, defines the URL used to generate the QR code.
    """

    new_field_spec: Optional[SigFieldSpec] = None
    """
    Field spec used to generate new signature fields, if applicable.
    """

    prefer_pss: bool = False
    """
    When working with RSA keys, prefer RSASSA-PSS signing if available.
    """

    detach_pem: bool = False
    """
    When producing detached signature payloads (i.e. non-PDF CMS), save the
    result in a PEM file instead of in a DER file.
    """

    lenient: bool = False
    """
    Process PDF files in nonstrict mode.
    """

    ux: UXContext = field(default_factory=UXContext)
    """
    UX information.
    """
